#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from __future__ import annotations

import os
import warnings

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"

import re

from pydantic import BaseModel, ConfigDict

import numpy as np
import pandas as pd

from copy import deepcopy as copy

import sklearn

sklearn.set_config(
    assume_finite=True, skip_parameter_validation=True
)  # Faster, we do checking

from scipy.sparse import csr_matrix

from scipy.spatial.distance import cdist

from sklearn.linear_model import ElasticNet, LinearRegression, Ridge, Lasso
from sklearn.kernel_ridge import KernelRidge
from sklearn.preprocessing import StandardScaler, RobustScaler

from timeit import default_timer as timer

import json

from opendsm.eemeter.models.hourly import settings as _settings
from opendsm.eemeter.models.hourly import HourlyBaselineData, HourlyReportingData
from opendsm.eemeter.common.exceptions import (
    DataSufficiencyError,
    DisqualifiedModelError,
)
from opendsm.eemeter.common.warnings import EEMeterWarning
from opendsm.common.clustering.cluster import cluster_features
from opendsm.common.adaptive_loss import adaptive_weights
from opendsm.common.metrics import BaselineMetrics, BaselineMetricsFromDict
from opendsm import __version__



class AdaptiveElasticNetRegressor:  
    def __init__(self, base_model, settings):
        self.settings = settings

        self.base_model = base_model
        self.base_model.warm_start = True

        self._hour_model = copy(self.base_model)

    
    def fit(self, X, y, sample_weight=None):
        """
        Fit the model with X, y data.
        
        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            The target values.
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
            
        Returns:
        --------
        self : returns an instance of self.
        """
        settings = self.settings.adaptive_weights
        window_size = self.settings.adaptive_weights.window_size - 1
        tol = self.settings.adaptive_weights.tol

        num_hours = y.shape[1]

        hour_model = copy(self.base_model)

        # fit the base model as an initial guess
        self.base_model.fit(X, y, sample_weight=sample_weight)

        if sample_weight is None:
            weights = np.ones((X.shape[0], num_hours))
        else:
            weights = sample_weight

        hour_fit = [False for _ in range(num_hours)]
        alpha_prior = np.array([2.0 for _ in range(num_hours)])
        for i in range(settings.max_iter):
            if all(hour_fit):
                i -= 1
                break

            # get prediction and residuals for all hours
            y_fit = self.base_model.predict(X)
            resid = y - y_fit

            for hour in range(num_hours):
                # if hour_fit[hour]:
                #     continue

                # Update weights
                # Calculate weights using window of hours
                window_idx = np.arange(hour - window_size, hour + window_size + 1)

                # if idx_i < 0, roll to the end or if idx_i >= num_hours, roll to the beginning 
                for idx_i in range(len(window_idx)):
                    if window_idx[idx_i] < 0:
                        window_idx[idx_i] = num_hours + window_idx[idx_i]

                    if window_idx[idx_i] >= num_hours:
                        window_idx[idx_i] = window_idx[idx_i] - num_hours

                # unique values in idx only
                window_idx = list(set(window_idx))
  
                # calculate weights
                weights_update, _, alpha = adaptive_weights(
                    resid[:,window_idx].flatten(), 
                    alpha="adaptive", 
                    sigma=settings.sigma, 
                    quantile=0.25, 
                    min_weight=0.0,
                    C_algo=settings.c_algo,
                )

                # break criteria
                if (alpha == 2) or (np.abs(alpha - alpha_prior[hour]) <= tol):
                    hour_fit[hour] = True
                    continue
                else:
                    hour_fit[hour] = False

                # update weights and alpha_prior
                alpha_prior[hour] = alpha

                # trim weights to hour size
                if window_size > 0:
                    # get index of hour in window_idx
                    idx = window_idx.index(hour)
                    hour_len = int(len(weights_update)/len(window_idx))

                    weights_update = weights_update[idx*hour_len:(idx+1)*hour_len]

                weights[:, hour] *= weights_update
                
                # update hour model from base model
                self._hour_model.coef_ = self.base_model.coef_[hour,:]
                self._hour_model.intercept_ = self.base_model.intercept_[hour]

                # fit
                self._hour_model.fit(
                    X, 
                    y[:, hour], 
                    sample_weight=weights[:, hour]
                )
                
                # update base model from refit hour model            
                self.base_model.coef_[hour,:] = self._hour_model.coef_
                self.base_model.intercept_[hour] = self._hour_model.intercept_

        # save info to base_model
        self.base_model.adaptive_iterations = i
        self.base_model.adaptive_alpha = alpha_prior
        self.base_model.adaptive_weights = weights
           
        return self

    @property
    def is_fit(self):
        """Check if the model is fitted."""
        is_fit = True

        if not hasattr(self.base_model, "coef_"):
            is_fit = False

        if not hasattr(self.base_model, "intercept_"):
            is_fit = False

        return is_fit
    
    def predict(self, X):
        """
        Predict using the model.
        
        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
            
        Returns:
        --------
        y : array of shape (n_samples,) or (n_samples, n_targets)
            The predicted values.
        """
        if not self.is_fit:
            raise RuntimeError("Model must be fit before predictions can be made.")

        y = self.base_model.predict(X)

        return y
    
    @property
    def coef_(self):
        """Get model coefficients."""
        if not hasattr(self.base_model, "coef_"):
            raise RuntimeError("Model coefficients must be set before accessed.")
        
        return self.base_model.coef_
        
    @coef_.setter
    def coef_(self, val):
        self.base_model.coef_ = val

    @property
    def intercept_(self):
        """Get model intercepts."""
        if not hasattr(self.base_model, "intercept_"):
            raise RuntimeError("Model intercepts must be set before accessed.")

        return self.base_model.intercept_
        
    @intercept_.setter
    def intercept_(self, val):
        """Set model intercepts"""
        self.base_model.intercept_ = val


class HourlyModel:
    """
    A class to fit a model to the input meter data.

    Attributes:
        settings (dict): A dictionary of settings.
        baseline_metrics (dict): A dictionary of metrics based on input baseline data and model fit.
    """
    
    # thresholds for switching model types
    _alpha_model_threshold = 1E-5
    _l1_ratio_model_threshold = 1E-3

    # set priority columns for sorting
    # this is critical for ensuring predict column order matches fit column order
    _priority_cols = {
        "ts": ["temporal_cluster", "temp_bin", "temperature", "ghi"],
        "cat": ["temporal_cluster", "temp_bin"],
    }

    _temporal_cluster_cols = ["month", "day_of_week"]

    """Note:
        Despite the temporal clusters, we can view all models created as a subset of the same full model.
        The temporal clusters would simply have the same coefficients within the same days/month combinations.
    """

    def __init__(
        self,
        settings: dict | _settings.BaseHourlySettings | None = None,
    ):
        """
        Args:
            settings: HourlySettings to use (generally left default). Will default to solar model if GHI is given to the fit step.
        """

        # TODO move this logic into HourlySettings init
        if isinstance(settings, dict):
            if features := settings.get("train_features"):
                if "ghi" in features:
                    settings = _settings.HourlySolarSettings(**settings)
                else:
                    settings = _settings.HourlyNonSolarSettings(**settings)
            else:
                settings = _settings.BaseHourlySettings(**settings)

        # Initialize settings
        if settings is None:
            self.settings = _settings.BaseHourlySettings()
        else:
            self.settings = settings

        # Initialize model
        self._set_scalers()
        self._set_model()

        self._T_bin_edges = None
        self._T_edge_bin_coeffs = None
        self._T_edge_bin_rate = None
        self._df_temporal_clusters = None
        self._categorical_features = None
        self._ts_feature_norm = None

        self._ts_features = []
        if self.settings.train_features:
            self._ts_features = self.settings.train_features.copy()

        self.is_fitted = False
        self.baseline_metrics = None

        self.warnings: list[EEMeterWarning] = []
        self.disqualification: list[EEMeterWarning] = []

        self.baseline_timezone = None
        self.error = dict()
        self.version = __version__

    
    def _set_scalers(self):
        # set scalers
        if self.settings.scaling_method == _settings.ScalingChoice.STANDARD_SCALER:
            self._feature_scaler = StandardScaler()
            self._y_scaler = StandardScaler()
        elif self.settings.scaling_method == _settings.ScalingChoice.ROBUST_SCALER:
            self._feature_scaler = RobustScaler(unit_variance=True)
            self._y_scaler = RobustScaler(unit_variance=True)


    def _set_model(self):
        # set base model
        if self.settings.base_model == _settings.BaseModel.ELASTICNET:
            settings = self.settings.elasticnet
            if settings.alpha <= self._alpha_model_threshold:
                self._model = LinearRegression(
                    fit_intercept=settings.fit_intercept
                )
            else:
                if settings.l1_ratio <= self._l1_ratio_model_threshold:
                    model = Ridge
                elif settings.l1_ratio >= (1 - self._l1_ratio_model_threshold):
                    model = Lasso
                else:
                    model = ElasticNet

                self._model = model(
                    alpha=settings.alpha,
                    fit_intercept=settings.fit_intercept,
                    precompute=settings.precompute,
                    max_iter=settings.max_iter,
                    tol=settings.tol,
                    selection=settings.selection,
                    warm_start=settings.warm_start,
                    random_state=settings._seed,
                )

                if model == ElasticNet:
                    self._model.l1_ratio = settings.l1_ratio

            if self.settings.adaptive_weights.enabled:
                self._model = AdaptiveElasticNetRegressor(self._model, self.settings)
                
        elif self.settings.base_model == _settings.BaseModel.KERNEL_RIDGE:
            settings = self.settings.kernel_ridge
            self._model = KernelRidge(
                alpha=settings.alpha,
                kernel=settings.kernel,
                gamma=settings.gamma,
            )


    def fit(
        self, baseline_data: HourlyBaselineData, ignore_disqualification: bool = False
    ) -> HourlyModel:
        """Fit the model using baseline data.

        Args:
            baseline_data: HourlyBaselineData object.
            ignore_disqualification: Whether to ignore disqualification errors / warnings.

        Returns:
            The fitted model.

        Raises:
            TypeError: If baseline_data is not an HourlyBaselineData object.
            DataSufficiencyError: If the model can't be fit on disqualified baseline data.
        """
        if not isinstance(baseline_data, HourlyBaselineData):
            raise TypeError("baseline_data must be an HourlyBaselineData object")
        baseline_data.log_warnings()
        if baseline_data.disqualification and not ignore_disqualification:
            raise DataSufficiencyError("Can't fit model on disqualified baseline data")
        if "ghi" in self._ts_features and not "ghi" in baseline_data.df.columns:
            raise ValueError(
                "Model was explicitly set to use GHI, but baseline data does not contain GHI."
            )

        self.warnings = baseline_data.warnings
        self.disqualification = baseline_data.disqualification

        if not self._ts_features:
            self.settings = self.settings.add_default_features(baseline_data.df.columns)
            self._ts_features = self.settings.train_features.copy()

        if "ghi" in baseline_data.df.columns and not "ghi" in self._ts_features:
            model_mismatch_warning = EEMeterWarning(
                qualified_name="eemeter.potential_model_mismatch",
                description=(
                    "Model was explicitly set to ignore GHI, but baseline period contained a GHI column."
                ),
                data={},
            )
            model_mismatch_warning.warn()
            self.warnings.append(model_mismatch_warning)

        self._fit(baseline_data)

        if not self._model_fit_is_acceptable():
            model_fit_warning = EEMeterWarning(
                qualified_name="eemeter.model_fit_metrics",
                description="Model disqualified due to poor fit.",
                data={
                    "cvrmse_threshold": self.settings.cvrmse_threshold,
                    "cvrmse_adj": self.baseline_metrics.cvrmse_adj,
                    "pnrmse_threshold": self.settings.pnrmse_threshold,
                    "pnrmse_adj": self.baseline_metrics.pnrmse_adj,
                },
            )
            model_fit_warning.warn()
            self.disqualification.append(model_fit_warning)
        return self

    def _fit(self, meter_data):
        self.is_fitted = False

        # Initialize dataframe
        df_meter = meter_data.df  # used to have a copy here

        # Prepare feature arrays/matrices
        X, y, fit_mask = self._prepare_features(df_meter)
        X_fit = X[fit_mask, :]
        y_fit = y[fit_mask]

        # fit the model
        self._model.fit(X_fit, y_fit)
        self.is_fitted = True

        # get number of model parameters
        if self.settings.base_model == _settings.BaseModel.ELASTICNET:
            if self.settings.adaptive_weights.enabled:
                self._model = self._model.base_model

            num_parameters = np.count_nonzero(self._model.coef_) + np.count_nonzero(
                self._model.intercept_
            )
        elif self.settings.base_model == _settings.BaseModel.KERNEL_RIDGE:
            num_parameters = np.count_nonzero(self._model.dual_coef_)
        
        # get model prediction of baseline
        df_meter = self._predict(meter_data, X=X)

        # calculate baseline metrics on non-interpolated data
        cols = [col for col in df_meter.columns if col.startswith("interpolated_")]
        interpolated = df_meter[cols].any(axis=1)

        self.baseline_metrics = BaselineMetrics(
            df=df_meter.loc[~interpolated], num_model_params=num_parameters
        )
        self.baseline_timezone = meter_data.tz

        return self

    def predict(
        self,
        reporting_data,
        ignore_disqualification=False,
    ) -> pd.DataFrame:
        """Predicts the energy consumption using the fitted model.

        Args:
            reporting_data (Union[HourlyBaselineData, HourlyReportingData]): The data used for prediction.
            ignore_disqualification (bool, optional): Whether to ignore model disqualification. Defaults to False.

        Returns:
            Dataframe with input data along with predicted energy consumption.

        Raises:
            RuntimeError: If the model is not fitted.
            DisqualifiedModelError: If the model is disqualified and ignore_disqualification is False.
            TypeError: If the reporting data is not of type HourlyBaselineData or HourlyReportingData.
        """
        if not self.is_fitted:
            raise RuntimeError("Model must be fit before predictions can be made.")

        if missing_features := (
            set(self._ts_features) - set(reporting_data.df.columns)
        ):
            raise ValueError(
                f"Reporting data is missing the following features: {missing_features}"
            )

        if "ghi" in reporting_data.df.columns and not "ghi" in self._ts_features:
            model_mismatch_warning = EEMeterWarning(
                qualified_name="eemeter.potential_model_mismatch",
                description=(
                    "Reporting data contains GHI, but model was fit without GHI."
                ),
                data={},
            )
            model_mismatch_warning.warn()
            self.warnings.append(model_mismatch_warning)

        if str(self.baseline_timezone) != str(reporting_data.tz):
            raise ValueError(
                "Reporting data must use the same timezone that the model was initially fit on."
            )

        if self.disqualification and not ignore_disqualification:
            raise DisqualifiedModelError(
                "Attempting to predict using disqualified model without setting ignore_disqualification=True"
            )

        if not isinstance(reporting_data, (HourlyBaselineData, HourlyReportingData)):
            raise TypeError(
                "reporting_data must be a HourlyBaselineData or HourlyReportingData object"
            )

        return self._predict(reporting_data)

    def _predict(self, eval_data, X=None):
        """
        Makes model prediction on given temperature data.

        Parameters:
            df_eval (pandas.DataFrame): The evaluation dataframe.

        Returns:
            pandas.DataFrame: The evaluation dataframe with model predictions added.
        """

        df_eval = eval_data.df  # used to have a copy here
        dst_indices = _get_dst_indices(df_eval)
        datetime_original = eval_data.df.index
        # # get list of columns to keep in output
        columns = df_eval.columns.tolist()
        if "datetime" in columns:
            columns.remove("datetime")  # index in output, not column

        if X is None:
            X, _, _ = self._prepare_features(df_eval)

        y_predict_scaled = self._model.predict(X)
        y_predict = self._y_scaler.inverse_transform(y_predict_scaled)
        y_predict = y_predict.flatten()

        y_predict = _transform_dst(y_predict, dst_indices)

        df_eval["predicted"] = y_predict

        # # remove columns not in original columns and predicted
        df_eval = df_eval[[*columns, "predicted"]]

        # reindex to original datetime index
        df_eval = df_eval.reindex(datetime_original)

        return df_eval

    def _prepare_features(self, meter_data):
        """
        Initializes the meter data by performing the following operations:
        - Renames the 'model' column to 'model_old' if it exists
        - Converts the index to a DatetimeIndex if it is not already
        - Adds a 'season' column based on the month of the index using the settings.season dictionary
        - Adds a 'day_of_week' column based on the day of the week of the index
        - Removes any rows with NaN values in the 'temperature' or 'observed' columns
        - Sorts the data by the index
        - Reorders the columns to have 'season' and 'day_of_week' first, followed by the remaining columns

        Parameters:
        - meter_data: A pandas DataFrame containing the meter data

        Returns:
        - A pandas DataFrame containing the initialized meter data
        """
        dst_indices = _get_dst_indices(meter_data)
        meter_data = self._add_categorical_features(meter_data)
        self._add_supplemental_features(meter_data)

        self._ts_features, self._categorical_features = self._sort_features(
            self._ts_features, self._categorical_features
        )

        meter_data = self._daily_fitting_sufficiency(meter_data)
        meter_data = self._normalize_features(meter_data)
        meter_data = self._add_temperature_interactions(meter_data)

        # save actual df used for later inspection
        self._ts_feature_norm, _ = self._sort_features(self._ts_feature_norm)
        selected_features = self._ts_feature_norm + self._categorical_features
        if "observed_norm" in meter_data.columns:
            selected_features += ["observed_norm"]
        self._processed_meter_data_full = meter_data
        self._processed_meter_data = self._processed_meter_data_full[selected_features]

        # get feature matrices
        X, y, fit_mask = self._get_feature_matrices(meter_data, dst_indices)

        # Convert to sparse matrix
        X = csr_matrix(X.astype(float))

        return X, y, fit_mask

    def _add_temperature_bins(self, df):
        # TODO: do we need to do something about empty bins in prediction? I think not but maybe
        settings = self.settings.temperature_bin

        # add temperature bins based on temperature
        if not self.is_fitted:
            if settings.method == "equal_sample_count":
                T_bin_edges = pd.qcut(
                    df["temperature"], q=settings.n_bins, labels=False
                )

            elif settings.method == "equal_bin_width":
                T_bin_edges = pd.cut(
                    df["temperature"], bins=settings.n_bins, labels=False
                )

            elif settings.method == "set_bin_width":
                bin_width = settings.bin_width

                min_temp = np.floor(df["temperature"].min())
                max_temp = np.ceil(df["temperature"].max())

                if not settings.include_edge_bins:
                    step_num = (
                        np.round((max_temp - min_temp) / bin_width).astype(int) + 1
                    )

                    # T_bin_edges = np.arange(min_temp, max_temp + bin_width, bin_width)
                    T_bin_edges = np.linspace(min_temp, max_temp, step_num)

                else:
                    set_edge_bin_width = False
                    if set_edge_bin_width:
                        edge_bin_width = bin_width * 1 / 2

                        bin_range = [
                            min_temp + edge_bin_width,
                            max_temp - edge_bin_width,
                        ]

                    else:
                        edge_bin_count = int(len(df) * settings.edge_bin_percent)

                        # get 5th smallest and 5th largest temperatures
                        sorted_temp = np.sort(df["temperature"])
                        min_temp_reg_bin = np.ceil(sorted_temp[edge_bin_count])
                        max_temp_reg_bin = np.floor(sorted_temp[-edge_bin_count])

                        bin_range = [min_temp_reg_bin, max_temp_reg_bin]

                    step_num = (
                        np.round((bin_range[1] - bin_range[0]) / bin_width).astype(int)
                        + 1
                    )

                    # create bins with set width
                    T_bin_edges = np.array(
                        [min_temp, *np.linspace(*bin_range, step_num), max_temp]
                    )

            elif settings.method == "fixed_bins":
                temp =  df["temperature"].values
                min_temp = np.floor(np.min(temp))
                max_temp = np.ceil(np.max(temp))

                T_bin_edges = np.array(settings.fixed_bins)
                T_bin_edges = np.array([-np.inf, *T_bin_edges, np.inf])

                # if less than 20 values from df["temperature"] are in a bin, remove bin edge starting from edges and moving inwards
                idx_remove = []

                # count from left
                for i in range(len(T_bin_edges) - 1):
                    bin_count = ((temp >= T_bin_edges[i]) & (temp < T_bin_edges[i + 1])).sum()
                    if bin_count < settings.min_bin_count:
                        idx_remove.append(i+1)
                    else:
                        break

                # count from right
                for i in range(len(T_bin_edges) - 1, 0, -1):
                    bin_count = ((temp >= T_bin_edges[i - 1]) & (temp < T_bin_edges[i])).sum()
                    if bin_count < settings.min_bin_count:
                        idx_remove.append(i - 1)
                    else:
                        break

                # remove idx_remove from T_bin_edges
                T_bin_edges = np.delete(T_bin_edges, idx_remove)

            else:
                raise ValueError("Invalid temperature binning method")

            # set the first and last bin to -inf and inf
            T_bin_edges[0] = -np.inf
            T_bin_edges[-1] = np.inf

            # store bin edges for prediction
            self._T_bin_edges = T_bin_edges

        T_bins = pd.cut(df["temperature"], bins=self._T_bin_edges, labels=False)

        df["temp_bin"] = T_bins

        # Create dummy variables for temperature bins
        bin_dummies = pd.get_dummies(
            pd.Categorical(
                df["temp_bin"], categories=range(len(self._T_bin_edges) - 1)
            ),
            prefix="temp_bin",
        )
        bin_dummies.index = df.index

        col_names = bin_dummies.columns.tolist()
        df = pd.merge(df, bin_dummies, how="left", left_index=True, right_index=True)

        return df, col_names

    def _add_categorical_features(self, df):
        def set_initial_temporal_clusters(df):
            fit_df_grouped = (
                df.groupby(self._temporal_cluster_cols + ["hour_of_day"])["observed"]
                .agg(self.settings.temporal_cluster_aggregation)
                .reset_index()
            )
            # pivot table to get 2D array of observed values
            fit_df_grouped = fit_df_grouped.pivot_table(
                index=self._temporal_cluster_cols,
                columns="hour_of_day",
                values="observed",
            )

            labels = cluster_features(
                fit_df_grouped,
                self.settings.temporal_cluster
            )

            df_temporal_clusters = pd.DataFrame(
                labels,
                columns=["temporal_cluster"],
                index=fit_df_grouped.index,
            )

            return df_temporal_clusters

        def correct_missing_temporal_clusters(df):
            # check and match any missing temporal combinations

            # get all unique combinations of month and day_of_week in df
            df_temporal = df[self._temporal_cluster_cols].drop_duplicates()
            df_temporal = df_temporal.sort_values(self._temporal_cluster_cols)
            df_temporal_index = df_temporal.set_index(self._temporal_cluster_cols).index

            # reindex self.df_temporal_clusters to df_temporal_index
            df_temporal_clusters = self._df_temporal_clusters.reindex(df_temporal_index)

            # get index of any nan values in df_temporal_clusters
            missing_combinations = df_temporal_clusters[
                df_temporal_clusters["temporal_cluster"].isna()
            ].index
            if not missing_combinations.empty:
                if "observed" in df.columns and not df["observed"].isnull().all():
                    # filter df to only include missing combinations
                    df_missing = df[
                        df.set_index(self._temporal_cluster_cols).index.isin(
                            missing_combinations
                        )
                    ]

                    df_missing_grouped = (
                        df_missing.groupby(
                            self._temporal_cluster_cols + ["hour_of_day"]
                        )["observed"]
                        .agg(self.settings.temporal_cluster_aggregation)
                        .reset_index()
                    )
                    df_missing_grouped = df_missing_grouped.pivot_table(
                        index=self._temporal_cluster_cols,
                        columns="hour_of_day",
                        values="observed",
                    )
                    X = df_missing_grouped.values

                    # calculate average observed for known clusters
                    # join df_temporal_clusters to df on month and day_of_week
                    df = pd.merge(
                        df,
                        df_temporal_clusters,
                        how="left",
                        left_on=self._temporal_cluster_cols,
                        right_index=True,
                    )

                    df_known = df[
                        ~df.set_index(self._temporal_cluster_cols).index.isin(
                            missing_combinations
                        )
                    ]

                    df_known_mean = (
                        df_known.groupby(self._temporal_cluster_cols + ["hour_of_day"])[
                            "observed"
                        ]
                        .mean()
                        .reset_index()
                    )
                    df_known_mean = df_known_mean.pivot_table(
                        index=self._temporal_cluster_cols,
                        columns="hour_of_day",
                        values="observed",
                    )
                    X_known = df_known_mean.values

                    # get smallest distance between X and X_known
                    dist = cdist(X, X_known, metric="euclidean")
                    min_dist_idx = np.argmin(dist, axis=1)

                    # get temporal clusters df_known
                    temporal_clusters = df_known.groupby(self._temporal_cluster_cols)[
                        "temporal_cluster"
                    ].first()
                    temporal_clusters = temporal_clusters.reindex(df_known_mean.index)

                    # set labels to minimum distance of known clusters
                    labels = temporal_clusters.iloc[min_dist_idx].values
                    df_temporal_clusters.loc[
                        missing_combinations, "temporal_cluster"
                    ] = labels

                    self._df_temporal_clusters = df_temporal_clusters

                else:
                    # TODO: There's better ways of handling this
                    # unstack and fill missing days in each month
                    # assuming months more important than days
                    df_temporal_clusters = df_temporal_clusters.unstack()

                    # fill missing days in each month
                    df_temporal_clusters = df_temporal_clusters.ffill(axis=1)
                    df_temporal_clusters = df_temporal_clusters.bfill(axis=1)

                    # fill missing months if any remaining empty
                    df_temporal_clusters = df_temporal_clusters.ffill(axis=0)
                    df_temporal_clusters = df_temporal_clusters.bfill(axis=0)

                    df_temporal_clusters = df_temporal_clusters.stack()

            return df_temporal_clusters

        # assign basic temporal features
        df["date"] = df.index.date
        df["month"] = df.index.month
        df["day_of_week"] = df.index.dayofweek
        df["hour_of_day"] = df.index.hour

        # assign temporal clusters
        if not self.is_fitted:
            self._df_temporal_clusters = set_initial_temporal_clusters(df)
            n_clusters = self._df_temporal_clusters["temporal_cluster"].nunique()

        else:
            self._df_temporal_clusters = correct_missing_temporal_clusters(df)

            # Get all unique temporal clusters from categorical features
            temporal_cluster = []
            for col in self._categorical_features:
                if "temporal_cluster" in col:
                    match = re.match(r'^temporal_cluster_(\d+)*', col)
                    if match and int(match.group(1)) not in temporal_cluster:
                        temporal_cluster.append(int(match.group(1)))

            n_clusters = len(temporal_cluster)

        # join df_temporal_clusters to df
        df = pd.merge(
            df,
            self._df_temporal_clusters,
            how="left",
            left_on=self._temporal_cluster_cols,
            right_index=True,
        )

        cluster_dummies = pd.get_dummies(
            pd.Categorical(df["temporal_cluster"], categories=range(n_clusters)),
            prefix="temporal_cluster",
        )
        cluster_dummies.index = df.index

        cluster_cat = [f"temporal_cluster_{i}" for i in range(n_clusters)]
        self._categorical_features = cluster_cat

        df = pd.merge(
            df, cluster_dummies, how="left", left_index=True, right_index=True
        )

        if self.settings.temperature_bin is not None:
            df, temp_bin_cols = self._add_temperature_bins(df)
            self._categorical_features.extend(temp_bin_cols)

        return df

    def _add_supplemental_features(self, df):
        # TODO: should either do upper or lower on all strs
        if self.settings.supplemental_time_series_columns is not None:
            for col in self.settings.supplemental_time_series_columns:
                if (col in df.columns) and (col not in self._ts_features):
                    self._ts_features.append(col)

        if self.settings.supplemental_categorical_columns is not None:
            for col in self.settings.supplemental_categorical_columns:
                if (
                    (col in df.columns)
                    and (col not in self._ts_features)
                    and (col not in self._categorical_features)
                ):
                    self._categorical_features.append(col)

    def _sort_features(self, ts_features=None, cat_features=None):
        features = {"ts": ts_features, "cat": cat_features}

        # sort features
        for _type in ["ts", "cat"]:
            feat = features[_type]

            if feat is not None:
                sorted_cols = []
                for col in self._priority_cols[_type]:
                    cat_cols = [c for c in feat if c.startswith(col)]
                    sorted_cols.extend(sorted(cat_cols))

                # get all columns in self._categorical_feature not in sorted_cat_cols
                leftover_cols = [c for c in feat if c not in sorted_cols]
                if leftover_cols:
                    sorted_cols.extend(sorted(leftover_cols))

                features[_type] = sorted_cols

        return features["ts"], features["cat"]

    # TODO rename to avoid confusion with data sufficiency
    def _daily_fitting_sufficiency(self, df):
        # remove days with insufficient data
        min_hours = self.settings.min_daily_training_hours

        if min_hours > 0:
            # find any rows with interpolated data
            cols = [col for col in df.columns if col.startswith("interpolated_")]
            df["interpolated"] = df[cols].any(axis=1)

            # if row contains any null values, set interpolated to True
            df["interpolated"] = df["interpolated"] | df.isnull().any(axis=1)

            # count number of non interpolated hours per day
            daily_hours = 24 - df.groupby("date")["interpolated"].sum()
            sufficient_days = daily_hours[daily_hours >= min_hours].index

            # set "include_day" column to True if day has sufficient hours
            df["include_date"] = df["date"].isin(sufficient_days)

        else:
            df["include_date"] = True

        return df

    def _normalize_features(self, df):
        """ """
        train_features = self._ts_features
        self._ts_feature_norm = [i + "_norm" for i in train_features]

        # need to set scaler if not fit
        if not self.is_fitted:
            self._feature_scaler.fit(df[train_features].values)
            self._y_scaler.fit(df["observed"].values.reshape(-1, 1))

        data_transformed = self._feature_scaler.transform(df[train_features].values)
        normalized_df = pd.DataFrame(
            data_transformed, index=df.index, columns=self._ts_feature_norm
        )

        df = pd.concat([df, normalized_df], axis=1)

        if "observed" in df.columns:
            df["observed_norm"] = self._y_scaler.transform(
                df["observed"].values.reshape(-1, 1)
            )

        return df
    
    def _add_extreme_temperature_bins(self, df, bin_range):
        settings = self.settings.temperature_bin

        def get_k(int_col, a, b):
            k = []
            for hour in range(24):
                df_hour = df[df["hour_of_day"] == hour]
                df_hour = df_hour.sort_values(by=int_col)

                x_data = a * df_hour[int_col].values + b
                y_data = df_hour["observed"].values

                # Fit the model using robust least squares
                try:
                    params = _fit_exp_growth_decay(
                        x_data, y_data, k_only=True, is_x_sorted=True
                    )
                    # save k for each hour
                    k.append(params[2])
                except:
                    pass

            k = np.abs(np.array(k))
            k_valid = k[k < 5]

            if len(k_valid) > 0:
                k = np.mean(k_valid)
            else:
                k = 1 # if no valid k, set to 1

            # if k is too small, set to minimum
            k_min = 1/np.log(1E6)
            if k < k_min:
                k = k_min

            return k

        if self._T_edge_bin_coeffs is None:
            self._T_edge_bin_coeffs = {}

        cols = bin_range
        # maybe add nonlinear terms to second and second to last columns?
        # cols = [0, 1, last_temp_bin - 1, last_temp_bin]
        # cols = list(set(cols))
        # all columns?
        # cols = range(cols[0], cols[1] + 1)

        # Add all columns using col_dict at end
        col_dict = {}
        for n in cols:
            base_col = f"temp_bin_{n}"
            int_col = f"{base_col}_ts"
            T_col = f"{base_col}_T"

            # get k for exponential growth/decay
            if not self.is_fitted:
                # determine temperature conversion for bin
                range_offset = settings.edge_bin_temperature_range_offset
                T_range = [
                    df[int_col].min() - range_offset,
                    df[int_col].max() + range_offset,
                ]
                new_range = [-1, 1]

                T_a = (new_range[1] - new_range[0]) / (T_range[1] - T_range[0])
                T_b = new_range[1] - T_a * T_range[1]

                # The best rate for exponential
                if settings.edge_bin_rate == "heuristic":
                    k = get_k(int_col, T_a, T_b)
                else:
                    k = settings.edge_bin_rate

                # get A for exponential
                A = 1 / (np.exp(1 / k * new_range[1]) - 1)

                self._T_edge_bin_coeffs[n] = {
                    "t_a": float(T_a),
                    "t_b": float(T_b),
                    "k": float(k),
                    "a": float(A),
                }

            T_a = self._T_edge_bin_coeffs[n]["t_a"]
            T_b = self._T_edge_bin_coeffs[n]["t_b"]
            k = self._T_edge_bin_coeffs[n]["k"]
            A = self._T_edge_bin_coeffs[n]["a"]

            col_dict[T_col] = np.where(
                df[base_col].values, T_a * df[int_col].values + T_b, 0
            )

            for pos_neg in ["pos", "neg"]:
                # if first or last column, add additional column
                # testing exp, previously squaring worked well

                s = 1
                if "neg" in pos_neg:
                    s = -1

                # set rate exponential
                ts_col = f"{base_col}_{pos_neg}_exp_ts"

                col_dict[ts_col] = np.where(
                    df[base_col].values, A * np.exp(s / k * col_dict[T_col]) - A, 0
                )

                self._ts_feature_norm.append(ts_col)

        # create new df with col_dict
        df = pd.concat([df, pd.DataFrame(col_dict, index=df.index)], axis=1)

        return df


    def _add_temperature_interactions(self, df):
        settings = self.settings.temperature_bin

        # TODO: if this permanent then it should not create, erase, make anew
        self._ts_feature_norm.remove("temperature_norm")

        temp_bin_cols = [c for c in df.columns if re.match(r'^temp_bin_\d+$', c)]
        cluster_cols = [c for c in df.columns if re.match(r'^temporal_cluster_\d+$', c)]

        col_dict = {}

        # add global temperature bins
        for col in temp_bin_cols:
            # splits temperature_norm into unique columns if that temp_bin column is True
            ts_col = f"{col}_ts"
            col_dict[ts_col] = df["temperature_norm"] * df[col]

            self._ts_feature_norm.append(ts_col)

        # add temporal cluster interactions
        # multiply each temp_bin by each temporal cluster
        # get all columns that start with temp_bin_ and are a number
        s = self.settings.interaction_scalar
        for temporal_cluster_col in cluster_cols:
            for temp_bin_col in temp_bin_cols:
                # add intercept term
                interaction_col = f"{temporal_cluster_col}_{temp_bin_col}_interact"
                col_dict[interaction_col] = df[temp_bin_col] * df[temporal_cluster_col]

                # add slope term
                interaction_ts_col = f"{interaction_col}_ts"
                # df[interaction_ts_col] = df["temperature_norm"] * df[interaction_col]
                col_dict[interaction_ts_col] = s*df["temperature_norm"] * col_dict[interaction_col]

                # add to feature lists
                self._categorical_features.append(interaction_col)
                self._ts_feature_norm.append(interaction_ts_col)

        # concat df with col_dict
        df = pd.concat([df, pd.DataFrame(col_dict, index=df.index)], axis=1)

        # TODO: Model is better without this, but not sure why
        # remove temporal cluster columns from categorical features
        # cluster_cols = [c for c in df.columns if re.match(r'^temporal_cluster_\d+(?!_)', c)]
        # self._categorical_features = [c for c in self._categorical_features if c not in cluster_cols]

        # add extreme temperature bins to global temperature bins
        if settings.include_edge_bins:
            bin_range = [0, len(temp_bin_cols) - 1]
            df = self._add_extreme_temperature_bins(df, bin_range)

        return df

    def _get_feature_matrices(self, df, dst_indices):
        # get aggregated features with agg function
        agg_dict = {f: lambda x: list(x) for f in self._ts_feature_norm}

        def correct_dst(agg):
            """interpolate or average hours to account for DST. modifies in place"""
            interp, mean = dst_indices
            for date, hour in interp:
                for feature_idx, feature in enumerate(agg[date]):
                    if hour == 0:
                        # there are a handful of countries that use 0:00 as the DST transition
                        interpolated = (
                            agg[date - 1][feature_idx][-1] + feature[hour]
                        ) / 2
                    else:
                        interpolated = (feature[hour - 1] + feature[hour]) / 2
                    feature.insert(hour, interpolated)
            for date, hour in mean:
                for feature in agg[date]:
                    mean = (feature[hour + 1] + feature.pop(hour)) / 2
                    feature[hour] = mean

        df_grouped = df.groupby("date")
        agg_x = df_grouped.agg(agg_dict).values.tolist()
        correct_dst(agg_x)

        # get the features and target for each day
        ts_feature = np.array(agg_x)

        ts_feature = ts_feature.reshape(
            ts_feature.shape[0], ts_feature.shape[1] * ts_feature.shape[2]
        )

        # get the first categorical features for each day for each sample
        unique_dummies = (
            df[["date"] + self._categorical_features].groupby("date").first()
        )

        X = np.concatenate((ts_feature, unique_dummies), axis=1)

        if not self.is_fitted:
            agg_y = (
                df_grouped
                .agg({"observed_norm": lambda x: list(x)})
                .values.tolist()
            )
            correct_dst(agg_y)
            y = np.array(agg_y)
            y = y.reshape(y.shape[0], y.shape[1] * y.shape[2])

            fit_mask = df_grouped["include_date"].first().values
        else:
            y = None
            fit_mask = None

        return X, y, fit_mask

    def _model_fit_is_acceptable(self):
        cvrmse = self.baseline_metrics.cvrmse_adj
        pnrmse = self.baseline_metrics.pnrmse_adj

        # sufficient is (0 <= cvrmse <= threshold) or (0 <= pnrmse <= threshold)

        if cvrmse is not None:
            if (0 <= cvrmse) and (cvrmse <= self.settings.cvrmse_threshold):
                return True
            
        if pnrmse is not None:
            # less than 0 is not possible, but just in case
            if (0 <= pnrmse) and (pnrmse <= self.settings.pnrmse_threshold):
                return True

        return False


    def to_dict(self) -> dict:
        """Returns a dictionary of model parameters.

        Returns:
            Model parameters.
        """
        feature_scaler = {}
        if self.settings.scaling_method == _settings.ScalingChoice.STANDARD_SCALER:
            for i, key in enumerate(self._ts_features):
                feature_scaler[key] = [
                    self._feature_scaler.mean_[i],
                    self._feature_scaler.scale_[i],
                ]

            y_scaler = [self._y_scaler.mean_.squeeze(), self._y_scaler.scale_.squeeze()]

        elif self.settings.scaling_method == _settings.ScalingChoice.ROBUST_SCALER:
            for i, key in enumerate(self._ts_features):
                feature_scaler[key] = [
                    self._feature_scaler.center_[i],
                    self._feature_scaler.scale_[i],
                ]

            y_scaler = [
                self._y_scaler.center_.squeeze(),
                self._y_scaler.scale_.squeeze(),
            ]

        # convert self._df_temporal_clusters to list of lists
        df_temporal_clusters = self._df_temporal_clusters.reset_index().values.tolist()

        params = _settings.SerializeModel(
            settings=self.settings,
            temporal_clusters=df_temporal_clusters,
            temperature_bin_edges=self._T_bin_edges,
            temperature_edge_bin_coefficients=self._T_edge_bin_coeffs,
            ts_features=self._ts_features,
            categorical_features=self._categorical_features,
            coefficients=self._model.coef_.tolist(),
            intercept=self._model.intercept_.tolist(),
            feature_scaler=feature_scaler,
            catagorical_scaler=None,
            y_scaler=y_scaler,
            baseline_metrics=self.baseline_metrics,
            info=_settings.ModelInfo(
                disqualification=self.disqualification,
                warnings=self.warnings,
                error=self.error,
                baseline_timezone=str(self.baseline_timezone),
                version=self.version,
            ),
        )

        model_dict = params.model_dump()
        return model_dict

    def to_json(self) -> str:
        """Returns a JSON string of model parameters.

        Returns:
            Model parameters.
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, data) -> HourlyModel:
        """Create a instance of the class from a dictionary (such as one produced from the to_dict method).

        Args:
            data (dict): The dictionary containing the model data.

        Returns:
            An instance of the class.
        """
        # get settings
        train_features = data.get("settings").get("train_features")

        if "ghi" in train_features:
            settings = _settings.HourlySolarSettings(**data.get("settings"))
        else:
            settings = _settings.HourlyNonSolarSettings(**data.get("settings"))

        # initialize model class
        model_cls = cls(settings=settings)

        df_temporal_clusters = pd.DataFrame(
            data.get("temporal_clusters"),
            columns=model_cls._temporal_cluster_cols + ["temporal_cluster"],
        ).set_index(model_cls._temporal_cluster_cols)

        model_cls._df_temporal_clusters = df_temporal_clusters
        model_cls._T_bin_edges = np.array(data.get("temperature_bin_edges"))
        model_cls._T_edge_bin_coeffs = {
            int(k): v for k, v in data.get("temperature_edge_bin_coefficients").items()
        }

        model_cls._ts_features = data.get("ts_features")
        model_cls._categorical_features = data.get("categorical_features")

        # set scalers
        feature_scaler_values = list(data.get("feature_scaler").values())
        feature_scaler_loc = [i[0] for i in feature_scaler_values]
        feature_scaler_scale = [i[1] for i in feature_scaler_values]

        y_scaler_values = data.get("y_scaler")

        if settings.scaling_method == _settings.ScalingChoice.STANDARD_SCALER:
            model_cls._feature_scaler.mean_ = np.array(feature_scaler_loc)
            model_cls._feature_scaler.scale_ = np.array(feature_scaler_scale)

            model_cls._y_scaler.mean_ = np.array(y_scaler_values[0])
            model_cls._y_scaler.scale_ = np.array(y_scaler_values[1])

        elif settings.scaling_method == _settings.ScalingChoice.ROBUST_SCALER:
            model_cls._feature_scaler.center_ = np.array(feature_scaler_loc)
            model_cls._feature_scaler.scale_ = np.array(feature_scaler_scale)

            model_cls._y_scaler.center_ = np.array(y_scaler_values[0])
            model_cls._y_scaler.scale_ = np.array(y_scaler_values[1])

        # set model
        model_cls._model.coef_ = np.array(data.get("coefficients"))
        model_cls._model.intercept_ = np.array(data.get("intercept"))

        model_cls.is_fitted = True

        # set baseline metrics
        model_cls.baseline_metrics = BaselineMetricsFromDict(
            data.get("baseline_metrics")
        )

        info = _settings.ModelInfo(**data.get("info"))
        model_cls.warnings = info.warnings
        model_cls.disqualification = info.disqualification
        model_cls.error = info.error
        model_cls.baseline_timezone = info.baseline_timezone
        model_cls.version = info.version

        return model_cls

    @classmethod
    def from_json(cls, str_data) -> HourlyModel:
        """Create an instance of the class from a JSON string.

        Args:
            str_data: The JSON string representing the object.

        Returns:
            An instance of the class.

        """
        return cls.from_dict(json.loads(str_data))

    def plot(
        self,
        df_eval: HourlyBaselineData | HourlyReportingData,
    ):
        """Plot a model fit with baseline or reporting data.

        Args:
            df_eval: The baseline or reporting data object to plot.
        """
        raise NotImplementedError


def _fit_exp_growth_decay(x, y, k_only=True, is_x_sorted=False):
    # Courtsey: https://math.stackexchange.com/questions/1337601/fit-exponential-with-constant
    #           https://www.scribd.com/doc/14674814/Regressions-et-equations-integrales
    #           Jean Jacquelin

    # fitting function is actual b*exp(c*x) + a

    # sort x in order
    x = np.array(x)
    y = np.array(y)
    n = len(x)

    if not is_x_sorted:
        sort_idx = np.argsort(x)
        x = x[sort_idx]
        y = y[sort_idx]

    s = [0]
    for i in range(1, len(x)):
        s.append(s[i - 1] + 0.5 * (y[i] + y[i - 1]) * (x[i] - x[i - 1]))

    s = np.array(s)

    x_diff_sq = np.sum((x - x[0]) ** 2)
    xs_diff = np.sum(s * (x - x[0]))
    s_sq = np.sum(s**2)
    xy_diff = np.sum((x - x[0]) * (y - y[0]))
    ys_diff = np.sum(s * (y - y[0]))

    A = np.array([[x_diff_sq, xs_diff], [xs_diff, s_sq]])
    b = np.array([xy_diff, ys_diff])

    _, c = np.linalg.solve(A, b)
    with np.errstate(divide='ignore'):
        k = 1 / c # ignore divide by zero, it will be filtered later

    if k_only:
        a, b = None, None
    else:
        theta_i = np.exp(c * x)

        theta = np.sum(theta_i)
        theta_sq = np.sum(theta_i**2)
        y_sum = np.sum(y)
        y_theta = np.sum(y * theta_i)

        A = np.array([[n, theta], [theta, theta_sq]])
        b = np.array([y_sum, y_theta])

        a, b = np.linalg.solve(A, b)

    return a, b, k


def _get_dst_indices(df):
    """
    given a datetime-indexed dataframe,
    return the indices which need to be interpolated and averaged
    in order to ensure exact 24 hour slots
    """
    # TODO test on baselines that begin/end on DST change
    counts = df.groupby(df.index.date).count()
    interp = counts[counts["observed"] == 23]
    mean = counts[counts["observed"] == 25]

    interp_idx = []
    for idx in interp.index:
        month = df.loc[idx.isoformat()]
        date_idx = counts.index.get_loc(idx)
        missing_hour = set(range(24)) - set(month.index.hour)
        if len(missing_hour) != 1:
            raise ValueError("too many missing hours")
        hour = missing_hour.pop()
        interp_idx.append((date_idx, hour))

    mean_idx = []
    for idx in mean.index:
        date_idx = counts.index.get_loc(idx)
        month = df.loc[idx.isoformat()]
        seen = set()
        for i in month.index:
            if i.hour in seen:
                hour = i.hour
                break
            seen.add(i.hour)
        mean_idx.append((date_idx, hour))

    return interp_idx, mean_idx


def _transform_dst(prediction, dst_indices):
    interp, mean = dst_indices

    START_END = 0
    REMOVE = 1
    INTERPOLATE = 2

    # get concrete indices
    remove_idx = [(REMOVE, date * 24 + hour) for date, hour in interp]
    interp_idx = [(INTERPOLATE, date * 24 + hour + 1) for date, hour in mean]

    # these values will be inserted for the 25th hour
    interpolated_vals = []
    for _, idx in interp_idx:
        interpolated = (prediction[idx - 1] + prediction[idx]) / 2
        interpolated_vals.append(interpolated)
    interpolation = iter(interpolated_vals)

    # sort "operations" by index (can't assume a strict back-and-forth ordering)
    ops = sorted(remove_idx + interp_idx, key=lambda t: t[1])

    # create fenceposts where slices end
    pairs = list(zip([(START_END, 0)] + ops, ops + [(START_END, None)]))
    slices = []
    for start, end in pairs:
        start_i = start[1]
        end_i = end[1]
        if start[0] == REMOVE:
            start_i += 1
        if start[0] == INTERPOLATE:
            slices.append([next(interpolation)])
        slices.append(prediction[slice(start_i, end_i)])
    return np.concatenate(slices)

    ## the block above is equivalent to:
    # shift = 0
    # for op in ops:
    #     if op[0] == REMOVE:
    #         # delete artificial DST hour
    #         idx = op[1] + shift
    #         prediction = np.delete(prediction, idx)
    #         shift -= 1
    #     if op[0] == INTERPOLATE:
    #         # interpolate missing DST hour
    #         idx = op[1] + shift
    #         interp = (prediction[idx - 1] + prediction[idx]) / 2
    #         prediction = np.insert(prediction, idx, interp)
    #         shift += 1
    # return prediction



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import pandas as pd
from scipy.stats import t

from opendsm.eemeter.common.warnings import EEMeterWarning

__all__ = ("ModelMetrics",)


def _compute_r_squared(combined):
    return combined[["predicted", "observed"]].corr().iloc[0, 1] ** 2


def _compute_r_squared_adj(r_squared, length, num_parameters):
    return 1 - (1 - r_squared) * (length - 1) / (length - num_parameters - 1)


def _compute_rmse(combined):
    return (combined["residuals"].astype(float) ** 2).mean() ** 0.5


def _compute_rmse_adj(combined, length, num_parameters):
    if length > num_parameters:
        return (
            (combined["residuals"].astype(float) ** 2).sum() / (length - num_parameters)
        ) ** 0.5
    else:
        return np.nan


def _compute_cvrmse(rmse, observed_mean):
    return rmse / observed_mean


def _compute_cvrmse_adj(rmse_adj, observed_mean):
    return rmse_adj / observed_mean


def _compute_mape(combined):
    return (combined["residuals"] / combined["observed"]).abs().mean()


def _compute_nmae(combined):
    return (combined["residuals"].astype(float).abs().sum()) / (
        combined["observed"].sum()
    )


def _compute_nmbe(combined):
    return combined["residuals"].astype(float).sum() / combined["observed"].sum()


def _compute_autocorr_resid(combined, autocorr_lags):
    return combined["residuals"].autocorr(lag=autocorr_lags)


def _json_safe_float(number):
    """
    JSON serialization for infinity can be problematic.
    See https://docs.python.org/2/library/json.html#basic-usage
    This function returns None if `number` is infinity or negative infinity.

    If the `number` cannot be converted to float, this will raise an exception.
    """
    if number is None:
        return None

    if isinstance(number, float):
        return None if np.isinf(number) or np.isnan(number) else number

    # errors if number is not float compatible
    return float(number)


class ModelMetricsFromJson(object):
    def __init__(
        self,
        observed_length,
        predicted_length,
        merged_length,
        num_parameters,
        observed_mean,
        predicted_mean,
        observed_variance,
        predicted_variance,
        observed_skew,
        predicted_skew,
        observed_kurtosis,
        predicted_kurtosis,
        observed_cvstd,
        predicted_cvstd,
        r_squared,
        r_squared_adj,
        rmse,
        rmse_adj,
        cvrmse,
        cvrmse_adj,
        mape,
        mape_no_zeros,
        num_meter_zeros,
        nmae,
        nmbe,
        autocorr_resid,
        confidence_level,
        n_prime,
        single_tailed_confidence_level,
        degrees_of_freedom,
        t_stat,
        cvrmse_auto_corr_correction,
        approx_factor_auto_corr_correction,
        fsu_base_term,
    ):
        self.observed_length = observed_length
        self.predicted_length = predicted_length
        self.merged_length = merged_length
        self.num_parameters = num_parameters
        self.observed_mean = observed_mean
        self.predicted_mean = predicted_mean
        self.observed_variance = observed_variance
        self.predicted_variance = predicted_variance
        self.observed_skew = observed_skew
        self.predicted_skew = predicted_skew
        self.observed_kurtosis = observed_kurtosis
        self.predicted_kurtosis = predicted_kurtosis
        self.observed_cvstd = observed_cvstd
        self.predicted_cvstd = predicted_cvstd
        self.r_squared = r_squared
        self.r_squared_adj = r_squared_adj
        self.rmse = rmse
        self.rmse_adj = rmse_adj
        self.cvrmse = cvrmse
        self.cvrmse_adj = cvrmse_adj
        self.mape = mape
        self.mape_no_zeros = mape_no_zeros
        self.num_meter_zeros = num_meter_zeros
        self.nmae = nmae
        self.nmbe = nmbe
        self.autocorr_resid = autocorr_resid
        self.confidence_level = confidence_level
        self.n_prime = n_prime
        self.single_tailed_confidence_level = single_tailed_confidence_level
        self.degrees_of_freedom = degrees_of_freedom
        self.t_stat = t_stat
        self.cvrmse_auto_corr_correction = cvrmse_auto_corr_correction
        self.approx_factor_auto_corr_correction = approx_factor_auto_corr_correction
        self.fsu_base_term = fsu_base_term


class ModelMetrics(object):
    """Contains measures of model fit and summary statistics on the input series.

    Parameters
    ----------
    observed_input : :any:`pandas.Series`
        Series with :any:`pandas.DatetimeIndex` with a set of electricity or
        gas meter values.
    predicted_input : :any:`pandas.Series`
        Series with :any:`pandas.DatetimeIndex` with a set of electricity or
        gas meter values.
    num_parameters : :any:`int`, optional
        The number of parameters (excluding the intercept) used in the
        regression from which the predictions were derived.
    autocorr_lags : :any:`int`, optional
        The number of lags to use when calculating the autocorrelation of the
        residuals.
    confidence_level : :any:`int`, optional
        Confidence level used in fractional savings uncertainty computations.

    Attributes
    ----------
    observed_length : :any:`int`
        The length of the observed_input series.
    predicted_length : :any:`int`
        The length of the predicted_input series.
    merged_length : :any:`int`
        The length of the dataframe resulting from the inner join of the
        observed_input series and the predicted_input series.
    observed_mean : :any:`float`
        The mean of the observed_input series.
    predicted_mean : :any:`float`
        The mean of the predicted_input series.
    observed_skew : :any:`float`
        The skew of the observed_input series.
    predicted_skew : :any:`float`
        The skew of the predicted_input series.
    observed_kurtosis : :any:`float`
        The excess kurtosis of the observed_input series.
    predicted_kurtosis : :any:`float`
        The excess kurtosis of the predicted_input series.
    observed_cvstd : :any:`float`
        The coefficient of standard deviation of the observed_input series.
    predicted_cvstd : :any:`float`
        The coefficient of standard deviation of the predicted_input series.
    r_squared : :any:`float`
        The r-squared of the model from which the predicted_input series was
        produced.
    r_squared_adj : :any:`float`
        The r-squared of the predicted_input series relative to the
        observed_input series, adjusted by the number of parameters in the model.
    cvrmse : :any:`float`
        The coefficient of variation (root-mean-squared error) of the
        predicted_input series relative to the observed_input series.
    cvrmse_adj : :any:`float`
        The coefficient of variation (root-mean-squared error) of the
        predicted_input series relative to the observed_input series, adjusted
        by the number of parameters in the model.
    mape : :any:`float`
        The mean absolute percent error of the predicted_input series relative
        to the observed_input series.
    mape_no_zeros : :any:`float`
        The mean absolute percent error of the predicted_input series relative
        to the observed_input series, with all time periods dropped where the
        observed_input series was not greater than zero.
    num_meter_zeros : :any:`int`
        The number of time periods for which the observed_input series was not
        greater than zero.
    nmae : :any:`float`
        The normalized mean absolute error of the predicted_input series
        relative to the observed_input series.
    nmbe : :any:`float`
        The normalized mean bias error of the predicted_input series relative
        to the observed_input series.
    autocorr_resid : :any:`float`
        The autocorrelation of the residuals (where the residuals equal the
        predicted_input series minus the observed_input series), measured
        using a number of lags equal to autocorr_lags.
    n_prime: :any:`float`
        The number of baseline inputs corrected for autocorrelation -- used
        in fractional savings uncertainty computation.
    single_tailed_confidence_level: :any:`float`
        The adjusted confidence level for use in single-sided tests.
    degrees_of_freedom: :any:`float
        Maxmimum number of independent variables which have the freedom to vary
    t_stat: :any:`float
        t-statistic, used for hypothesis testing
    cvrmse_auto_corr_correction: :any:`float
        Correctoin factor the apply to cvrmse to account for autocorrelation of inputs.
    approx_factor_auto_corr_correction: :any:`float
        Approximation factor used in ashrae 14 guideline for uncertainty computation.
    fsu_base_term: :any:`float
        Base term used in fractional savings uncertainty computation.

    """

    def __init__(
        self,
        observed_input,
        predicted_input,
        num_parameters=1,
        autocorr_lags=1,
        confidence_level=0.90,
    ):
        if num_parameters < 0:
            raise ValueError("num_parameters must be greater than or equal to zero")
        if autocorr_lags <= 0:
            raise ValueError("autocorr_lags must be greater than zero")
        if (confidence_level <= 0) or (confidence_level >= 1):
            raise ValueError("confidence_level must be between zero and one.")

        self.warnings = []

        observed = observed_input.to_frame().dropna()
        predicted = predicted_input.to_frame().dropna()
        observed.columns = ["observed"]
        predicted.columns = ["predicted"]

        self.observed_length = observed.shape[0]
        self.predicted_length = predicted.shape[0]

        # Do an inner join on the two input series to make sure that we only
        # use observations with the same time stamps.
        combined = observed.merge(predicted, left_index=True, right_index=True)
        self.merged_length = len(combined)

        if self.observed_length != self.predicted_length:
            self.warnings.append(
                EEMeterWarning(
                    qualified_name="eemeter.metrics.input_series_are_of_different_lengths",
                    description="Input series are of different lengths.",
                    data={
                        "observed_input_length": len(observed_input),
                        "predicted_input_length": len(predicted_input),
                        "observed_length_without_nan": self.observed_length,
                        "predicted_length_without_nan": self.predicted_length,
                        "merged_length": self.merged_length,
                    },
                )
            )

        # Calculate residuals because these are an input for most of the metrics.
        combined["residuals"] = combined.predicted - combined.observed

        self.num_parameters = num_parameters
        self.autocorr_lags = autocorr_lags

        # to account for solar usage the cvrmse should be calculated as the
        # mean of the absolute value of observed.

        self.observed_mean = abs(combined["observed"]).mean()
        self.predicted_mean = abs(combined["predicted"]).mean()

        self.observed_variance = combined["observed"].var(ddof=0)
        self.predicted_variance = combined["predicted"].var(ddof=0)

        self.observed_skew = combined["observed"].skew()
        self.predicted_skew = combined["predicted"].skew()

        self.observed_kurtosis = combined["observed"].kurtosis()
        self.predicted_kurtosis = combined["predicted"].kurtosis()

        self.r_squared = _compute_r_squared(combined)
        self.r_squared_adj = _compute_r_squared_adj(
            self.r_squared, self.merged_length, self.num_parameters
        )

        self.rmse = _compute_rmse(combined)
        self.rmse_adj = _compute_rmse_adj(
            combined, self.merged_length, self.num_parameters
        )

        with np.errstate(divide="ignore", invalid="ignore"):
            self.observed_cvstd = combined["observed"].std() / self.observed_mean
            self.predicted_cvstd = combined["predicted"].std() / self.predicted_mean

            self.cvrmse = _compute_cvrmse(self.rmse, self.observed_mean)
            self.cvrmse_adj = _compute_cvrmse_adj(self.rmse_adj, self.observed_mean)

        # Create a new DataFrame with all rows removed where observed is
        # zero, so we can calculate a version of MAPE with the zeros excluded.
        # (Without the zeros excluded, MAPE becomes infinite when one observed
        # value is zero.)
        no_observed_zeros = combined[combined["observed"] > 0]

        self.mape = _compute_mape(combined)
        self.mape_no_zeros = _compute_mape(no_observed_zeros)

        self.num_meter_zeros = (self.merged_length) - no_observed_zeros.shape[0]

        self.nmae = _compute_nmae(combined)

        self.nmbe = _compute_nmbe(combined)

        self.autocorr_resid = _compute_autocorr_resid(combined, autocorr_lags)

        # ** Compute terms needed for fractional savings uncertainty computation.

        self.confidence_level = confidence_level
        self.n_prime = float(
            self.observed_length * (1 - self.autocorr_resid) / (1 + self.autocorr_resid)
        )
        self.single_tailed_confidence_level = 1 - ((1 - self.confidence_level) / 2)

        # convert to integer degrees of freedom, because n_prime could be non-integer
        if pd.isnull(self.n_prime) or not np.isfinite(self.n_prime):
            self.degrees_of_freedom = None
            self.t_stat = None
        else:
            self.degrees_of_freedom = round(self.n_prime - self.num_parameters)
            self.t_stat = t.ppf(
                self.single_tailed_confidence_level, self.degrees_of_freedom
            )

        if (
            self.n_prime == 0
            or pd.isnull(self.n_prime)
            or not np.isfinite(self.n_prime)
            or self.n_prime - self.num_parameters == 0
            or self.degrees_of_freedom < 1
            or self.observed_length < self.num_parameters
        ):
            self.cvrmse_auto_corr_correction = None
            self.approx_factor_auto_corr_correction = None
            self.fsu_base_term = None
        else:
            # factor to correct cvrmse_adj for autocorrelation of inputs
            # i.e., divide by (n' - n_param) instead of by (n - n_param)
            self.cvrmse_auto_corr_correction = (
                (self.observed_length - self.num_parameters)
                / (self.n_prime - self.num_parameters)
            ) ** 0.5

            # part of approximation factor used in ashrae 14 guideline
            self.approx_factor_auto_corr_correction = (
                1.0 + (2.0 / self.n_prime)
            ) ** 0.5

            # all the following values are unitless
            self.fsu_base_term = (
                self.t_stat
                * self.cvrmse_adj
                * self.cvrmse_auto_corr_correction
                * self.approx_factor_auto_corr_correction
            )

    def __repr__(self):
        return (
            "ModelMetrics(merged_length={}, r_squared_adj={}, cvrmse_adj={}, "
            "mape_no_zeros={}, nmae={}, nmbe={}, autocorr_resid={}, confidence_level={})".format(
                self.merged_length,
                round(self.r_squared_adj, 3),
                round(self.cvrmse_adj, 3),
                round(self.mape_no_zeros, 3),
                round(self.nmae, 3),
                round(self.nmbe, 3),
                round(self.autocorr_resid, 3),
                round(self.confidence_level, 3),
            )
        )

    def json(self):
        """Return a JSON-serializable representation of this result.

        The output of this function can be converted to a serialized string
        with :any:`json.dumps`.
        """
        return {
            "observed_length": _json_safe_float(self.observed_length),
            "predicted_length": _json_safe_float(self.predicted_length),
            "merged_length": _json_safe_float(self.merged_length),
            "num_parameters": _json_safe_float(self.num_parameters),
            "observed_mean": _json_safe_float(self.observed_mean),
            "predicted_mean": _json_safe_float(self.predicted_mean),
            "observed_variance": _json_safe_float(self.observed_variance),
            "predicted_variance": _json_safe_float(self.predicted_variance),
            "observed_skew": _json_safe_float(self.observed_skew),
            "predicted_skew": _json_safe_float(self.predicted_skew),
            "observed_kurtosis": _json_safe_float(self.observed_kurtosis),
            "predicted_kurtosis": _json_safe_float(self.predicted_kurtosis),
            "observed_cvstd": _json_safe_float(self.observed_cvstd),
            "predicted_cvstd": _json_safe_float(self.predicted_cvstd),
            "r_squared": _json_safe_float(self.r_squared),
            "r_squared_adj": _json_safe_float(self.r_squared_adj),
            "rmse": _json_safe_float(self.rmse),
            "rmse_adj": _json_safe_float(self.rmse_adj),
            "cvrmse": _json_safe_float(self.cvrmse),
            "cvrmse_adj": _json_safe_float(self.cvrmse_adj),
            "mape": _json_safe_float(self.mape),
            "mape_no_zeros": _json_safe_float(self.mape_no_zeros),
            "num_meter_zeros": _json_safe_float(self.num_meter_zeros),
            "nmae": _json_safe_float(self.nmae),
            "nmbe": _json_safe_float(self.nmbe),
            "autocorr_resid": _json_safe_float(self.autocorr_resid),
            "confidence_level": _json_safe_float(self.confidence_level),
            "n_prime": _json_safe_float(self.n_prime),
            "single_tailed_confidence_level": _json_safe_float(
                self.single_tailed_confidence_level
            ),
            "degrees_of_freedom": _json_safe_float(self.degrees_of_freedom),
            "t_stat": _json_safe_float(self.t_stat),
            "cvrmse_auto_corr_correction": _json_safe_float(
                self.cvrmse_auto_corr_correction
            ),
            "approx_factor_auto_corr_correction": _json_safe_float(
                self.approx_factor_auto_corr_correction
            ),
            "fsu_base_term": _json_safe_float(self.fsu_base_term),
        }

    @classmethod
    def from_json(cls, data):
        """Loads a JSON-serializable representation into the model state.

        The input of this function is a dict which can be the result
        of :any:`json.loads`.
        """

        c = ModelMetricsFromJson(
            observed_length=data.get("observed_length"),
            predicted_length=data.get("predicted_length"),
            merged_length=data.get("merged_length"),
            num_parameters=data.get("num_parameters"),
            observed_mean=data.get("observed_mean"),
            predicted_mean=data.get("predicted_mean"),
            observed_variance=data.get("observed_variance"),
            predicted_variance=data.get("predicted_variance"),
            observed_skew=data.get("observed_skew"),
            predicted_skew=data.get("predicted_skew"),
            observed_kurtosis=data.get("observed_kurtosis"),
            predicted_kurtosis=data.get("predicted_kurtosis"),
            observed_cvstd=data.get("observed_cvstd"),
            predicted_cvstd=data.get("predicted_cvstd"),
            r_squared=data.get("r_squared"),
            r_squared_adj=data.get("r_squared_adj"),
            rmse=data.get("rmse"),
            rmse_adj=data.get("rmse_adj"),
            cvrmse=data.get("cvrmse"),
            cvrmse_adj=data.get("cvrmse_adj"),
            mape=data.get("mape"),
            mape_no_zeros=data.get("mape_no_zeros"),
            num_meter_zeros=data.get("num_meter_zeros"),
            nmae=data.get("nmae"),
            nmbe=data.get("nmbe"),
            autocorr_resid=data.get("autocorr_resid"),
            confidence_level=data.get("confidence_level"),
            n_prime=data.get("n_prime"),
            single_tailed_confidence_level=data.get("single_tailed_confidence_level"),
            degrees_of_freedom=data.get("degrees_of_freedom"),
            t_stat=data.get("t_stat"),
            cvrmse_auto_corr_correction=data.get("cvrmse_auto_corr_correction"),
            approx_factor_auto_corr_correction=data.get(
                "approx_factor_auto_corr_correction"
            ),
            fsu_base_term=data.get("fsu_base_term"),
        )

        return c



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from __future__ import annotations

import pydantic

from enum import Enum
from typing import Optional

from opendsm.common.base_settings import BaseSettings, CustomField


class AlgorithmChoice(str, Enum):
    # SciPy scalar optimization algorithms
    SCIPY_BRENT = "scipy_brent"
    SCIPY_BOUNDED = "scipy_bounded"
    SCIPY_GOLDEN = "scipy_golden"

    # SciPy local optimization algorithms
    SCIPY_NELDERMEAD = "scipy_nelder-mead"
    SCIPY_L_BFGS_B = "scipy_l-bfgs-b"
    SCIPY_TNC = "scipy_tnc"
    SCIPY_COBYLA = "scipy_cobyla"
    SCIPY_COBYQA = "scipy_cobyqa"
    SCIPY_SLSQP = "scipy_slsqp"
    SCIPY_POWELL = "scipy_powell"
    SCIPY_TRUST_CONSTR = "scipy_trust-constr"

    # SciPy global optimization algorithms
    SCIPY_DIRECT = "scipy_direct"

    # nlopt-based algorithms
    NLOPT_DIRECT = "nlopt_direct"
    NLOPT_DIRECT_NOSCAL = "nlopt_direct_noscal"
    NLOPT_DIRECT_L = "nlopt_direct_l"
    NLOPT_DIRECT_L_RAND = "nlopt_direct_l_rand"
    NLOPT_DIRECT_L_NOSCAL = "nlopt_direct_l_noscal"
    NLOPT_DIRECT_L_RAND_NOSCAL = "nlopt_direct_l_rand_noscal"
    NLOPT_ORIG_DIRECT = "nlopt_orig_direct"
    NLOPT_ORIG_DIRECT_L = "nlopt_orig_direct_l"
    NLOPT_CRS2_LM = "nlopt_crs2_lm"
    NLOPT_MLSL_LDS = "nlopt_mlsl_lds"
    NLOPT_MLSL = "nlopt_mlsl"
    NLOPT_STOGO = "nlopt_stogo"
    NLOPT_STOGO_RAND = "nlopt_stogo_rand"
    NLOPT_AGS = "nlopt_ags"
    NLOPT_ISRES = "nlopt_isres"
    NLOPT_ESCH = "nlopt_esch"
    NLOPT_COBYLA = "nlopt_cobyla"
    NLOPT_BOBYQA = "nlopt_bobyqa"
    NLOPT_NEWUOA = "nlopt_newuoa"
    NLOPT_NEWUOA_BOUND = "nlopt_newuoa_bound"
    NLOPT_PRAXIS = "nlopt_praxis"
    NLOPT_NELDERMEAD = "nlopt_neldermead"
    NLOPT_SBPLX = "nlopt_sbplx"
    NLOPT_MMA = "nlopt_mma"
    NLOPT_CCSAQ = "nlopt_ccsaq"
    NLOPT_SLSQP = "nlopt_slsqp"
    NLOPT_L_BFGS = "nlopt_lbfgs"
    NLOPT_TNEWTON = "nlopt_tnewton"
    NLOPT_TNEWTON_PRECOND = "nlopt_tnewton_precond"
    NLOPT_TNEWTON_RESTART = "nlopt_tnewton_restart"
    NLOPT_TNEWTON_PRECOND_RESTART = "nlopt_tnewton_precond_restart"
    NLOPT_VAR1 = "nlopt_var1"
    NLOPT_VAR2 = "nlopt_var2"


class StopCriteriaChoice(str, Enum):
    ITERATION_MAXIMUM = "iteration maximum"
    MAXIMUM_TIME = "maximum time [min]"


class OptimizationSettings(BaseSettings):
    algorithm: AlgorithmChoice = CustomField(
        default=AlgorithmChoice.NLOPT_SBPLX,
        description="Optimization algorithm choice",
    )

    stop_criteria_type: StopCriteriaChoice = CustomField(
        default=StopCriteriaChoice.ITERATION_MAXIMUM,
        description="Stopping criteria",
    )

    stop_criteria_value: float = CustomField(
        default=2000,
        gt=0,
        description="Stopping criteria value for the optimization algorithm",
    )
    
    initial_step: Optional[float] = CustomField(
        default=0.1,
        description="Initial step size for the optimization algorithm",
    )

    x_tol_rel: float = CustomField(
        default=1e-5,
        gt=0,
        description="Relative cutoff X tolerance for the optimization algorithm",
    )

    f_tol_rel: float = CustomField(
        default=1e-5,
        gt=0,
        description="Relative cutoff function tolerance for the optimization algorithm",
    )

    initial_population_multiplier: Optional[float] = CustomField(
        default=None,
        description="Initial population multiplier for the optimization algorithm",
    )


    @pydantic.model_validator(mode="after")
    def _check_population_multiplier(self):
        if self.initial_population_multiplier is None:
            if self.algorithm == AlgorithmChoice.NLOPT_ISRES:
                raise ValueError("INITIAL_POPULATION_MULTIPLIER must be > 1 for nlopt_ISRES")
        else:
            if self.algorithm == AlgorithmChoice.NLOPT_ISRES:
                if self.initial_population_multiplier <= 1:
                    raise ValueError("INITIAL_POPULATION_MULTIPLIER must be > 1 for nlopt_ISRES")
            else:
                raise ValueError("INITIAL_POPULATION_MULTIPLIER must be None for all algorithms except nlopt_ISRES")

        return self


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pytest

from opendsm.eemeter.models.hourly_caltrack.design_matrices import (
    create_caltrack_hourly_preliminary_design_matrix,
    create_caltrack_hourly_segmented_design_matrices,
    create_caltrack_daily_design_matrix,
    create_caltrack_billing_design_matrix,
)
from opendsm.eemeter.common.features import (
    estimate_hour_of_week_occupancy,
    fit_temperature_bins,
)
from opendsm.eemeter.models.hourly_caltrack.segmentation import segment_time_series


def test_create_caltrack_hourly_preliminary_design_matrix(
    il_electricity_cdd_hdd_hourly,
):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    design_matrix = create_caltrack_hourly_preliminary_design_matrix(
        meter_data[:1000], temperature_data
    )
    assert design_matrix.shape == (1000, 7)
    assert sorted(design_matrix.columns) == [
        "cdd_65",
        "hdd_50",
        "hour_of_week",
        "meter_value",
        "n_hours_dropped",
        "n_hours_kept",
        "temperature_mean",
    ]
    # In newer pandas, categorical columns (like hour_of_week) arent included in sum
    design_matrix.hour_of_week = design_matrix.hour_of_week.astype(float)
    assert round(design_matrix.sum().sum(), 2) == 136352.61


def test_create_caltrack_daily_design_matrix(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    design_matrix = create_caltrack_daily_design_matrix(
        meter_data[:100], temperature_data
    )
    assert design_matrix.shape == (100, 6)
    assert sorted(design_matrix.columns) == [
        "meter_value",
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
        "temperature_not_null",
        "temperature_null",
    ]
    assert round(design_matrix.sum().sum(), 2) == 9267.06


def test_create_caltrack_billing_design_matrix(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    design_matrix = create_caltrack_billing_design_matrix(
        meter_data[:10], temperature_data
    )
    assert design_matrix.shape == (275, 6)
    assert sorted(design_matrix.columns) == [
        "meter_value",
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
        "temperature_not_null",
        "temperature_null",
    ]
    assert round(design_matrix.sum().sum(), 2) == 29925.27


@pytest.fixture
def preliminary_hourly_design_matrix(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    return create_caltrack_hourly_preliminary_design_matrix(
        meter_data[:1000], temperature_data
    )


@pytest.fixture
def segmentation(preliminary_hourly_design_matrix):
    return segment_time_series(
        preliminary_hourly_design_matrix.index, "three_month_weighted"
    )


@pytest.fixture
def occupancy_lookup(preliminary_hourly_design_matrix, segmentation):
    return estimate_hour_of_week_occupancy(
        preliminary_hourly_design_matrix, segmentation=segmentation
    )


@pytest.fixture
def temperature_bins(preliminary_hourly_design_matrix, segmentation, occupancy_lookup):
    return fit_temperature_bins(
        preliminary_hourly_design_matrix,
        segmentation=segmentation,
        occupancy_lookup=occupancy_lookup,
    )


def test_create_caltrack_hourly_segmented_design_matrices(
    preliminary_hourly_design_matrix, segmentation, occupancy_lookup, temperature_bins
):
    occupied_temperature_bins, unoccupied_temperature_bins = temperature_bins
    design_matrices = create_caltrack_hourly_segmented_design_matrices(
        preliminary_hourly_design_matrix,
        segmentation,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
    )

    design_matrix = design_matrices["dec-jan-feb-weighted"]
    assert design_matrix.shape == (1000, 8)
    assert sorted(design_matrix.columns) == [
        "bin_0_occupied",
        "bin_0_unoccupied",
        "bin_1_unoccupied",
        "bin_2_unoccupied",
        "bin_3_unoccupied",
        "hour_of_week",
        "meter_value",
        "weight",
    ]
    design_matrix.hour_of_week = design_matrix.hour_of_week.astype(float)
    assert round(design_matrix.sum().sum(), 2) == 126210.07

    design_matrix = design_matrices["mar-apr-may-weighted"]
    assert design_matrix.shape == (1000, 5)
    assert sorted(design_matrix.columns) == [
        "bin_0_occupied",
        "bin_0_unoccupied",
        "hour_of_week",
        "meter_value",
        "weight",
    ]
    design_matrix.hour_of_week = design_matrix.hour_of_week.astype(float)
    assert round(design_matrix.sum().sum(), 2) == 167659.28


def test_create_caltrack_billing_design_matrix_empty_temp(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"][:0]
    with pytest.raises(ValueError):
        design_matrix = create_caltrack_billing_design_matrix(
            meter_data[:10], temperature_data
        )


def test_create_caltrack_billing_design_matrix_partial_empty_temp(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"][:200]
    design_matrix = create_caltrack_billing_design_matrix(
        meter_data[:10], temperature_data
    )
    assert "n_days_kept" in design_matrix.columns



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from __future__ import annotations

from copy import deepcopy as copy
import warnings

import numpy as np
import scipy.sparse as sp

from sklearn.cluster import BisectingKMeans as _sklearn_BisectingKMeans
from sklearn.cluster import _bisect_k_means
from sklearn.cluster._kmeans import (
    _kmeans_single_elkan,
    _kmeans_single_lloyd,
)  # type: ignore
from sklearn.cluster._k_means_common import (
    _inertia_dense,
    _inertia_sparse,
)  # type: ignore
from sklearn.utils.extmath import row_norms
from sklearn.utils.validation import (
    _check_sample_weight,
    check_random_state,
)  # type: ignore

# from sklearn.utils._openmp_helpers import _openmp_effective_n_threads

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class BisectingKMeans(_sklearn_BisectingKMeans):
    """
    Override of sklearn class which simply saves the labels
    of all intermediate cluster steps.

    Only overrides fit

    Should always take the upper bound of number of clusters to try.
    Contains a new property named labels_full which is a dictionary where the key is the number of clusters
    and the value is the labels using that number.

    This should be used to score all the labels and determine the best number/labels to use/


    """

    def fit(self, X, y=None, sample_weight=None):
        """Compute bisecting k-means clustering.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)

            Training instances to cluster.

            .. note:: The data will be converted to C ordering,
                which will cause a memory copy
                if the given data is not C-contiguous.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        self
            Fitted estimator.
        """
        # return self._fit_test(X, y, sample_weight)

        self._validate_params()  # type: ignore

        with warnings.catch_warnings():
            # TODO deprecated in sklearn 1.6, removed in 1.7 (April 2025)
            warnings.filterwarnings("ignore", "`BaseEstimator._validate_data`")

            X = self._validate_data(  # type: ignore
                X,
                accept_sparse="csr",
                dtype=[np.float64, np.float32],
                order="C",
                copy=self.copy_x,  # type: ignore
                accept_large_sparse=False,
            )

        self._check_params_vs_input(X)  # type: ignore

        self._random_state = check_random_state(self.random_state)  # type: ignore
        sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)
        # self._n_threads = _openmp_effective_n_threads()
        self._n_threads = 1  # OVERRIDE OF ABOVE SO THAT RESULTS ARE DETERMINISTIC

        if self.algorithm == "lloyd" or self.n_clusters == 1:  # type: ignore
            self._kmeans_single = _kmeans_single_lloyd
            self._check_mkl_vcomp(X, X.shape[0])  # type: ignore
        else:
            self._kmeans_single = _kmeans_single_elkan

        # Subtract of mean of X for more accurate distance computations
        if not sp.issparse(X):
            self._X_mean = X.mean(axis=0)
            X -= self._X_mean

        # Initialize the hierarchical clusters tree
        self._bisecting_tree = _bisect_k_means._BisectingTree(
            indices=np.arange(X.shape[0]),
            center=X.mean(axis=0),
            score=0,
        )

        x_squared_norms = row_norms(X, squared=True)
        self.labels_full = {}
        for i in range(self.n_clusters - 1):  # type: ignore
            # Chose cluster to bisect
            try:
                cluster_to_bisect = self._bisecting_tree.get_cluster_to_bisect()
            except RecursionError:
                logger.warn(
                    f"encountered Recursion error during bisection for cluster size {i + 2}. Returning early"
                )
                return self

            # Split this cluster into 2 subclusters
            try:
                self._bisect(X, x_squared_norms, sample_weight, cluster_to_bisect)  # type: ignore
            except IndexError:
                logger.warn(
                    f"encountered IndexError during bisection for cluster size {i + 2}"
                )
                return self  # return early so that calculated labels can be returned until an error arose

            # Aggregate final labels and centers from the bisecting tree
            labels = np.full(X.shape[0], -1, dtype=np.int32)

            for j, cluster_node in enumerate(self._bisecting_tree.iter_leaves()):
                labels[cluster_node.indices] = j  # type: ignore

            self.labels_full[i + 2] = copy(labels)

        # Aggregate final labels and centers from the bisecting tree
        self.labels_ = np.full(X.shape[0], -1, dtype=np.int32)
        self.cluster_centers_ = np.empty((self.n_clusters, X.shape[1]), dtype=X.dtype)  # type: ignore

        for i, cluster_node in enumerate(self._bisecting_tree.iter_leaves()):
            self.labels_[cluster_node.indices] = i  # type: ignore
            self.cluster_centers_[i] = cluster_node.center  # type: ignore
            cluster_node.label = i  # type: ignore
            cluster_node.indices = None  # type: ignore

        # Restore original data
        if not sp.issparse(X):
            X += self._X_mean
            self.cluster_centers_ += self._X_mean

        _inertia = _inertia_sparse if sp.issparse(X) else _inertia_dense
        self.inertia_ = _inertia(
            X, sample_weight, self.cluster_centers_, self.labels_, self._n_threads
        )

        self._n_features_out = self.cluster_centers_.shape[0]

        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from math import ceil
from typing import Optional

import numpy as np
import pandas as pd
import pytz
from pandas.tseries.offsets import MonthBegin, MonthEnd

from opendsm.eemeter.common.warnings import EEMeterWarning


def remove_duplicates(df_or_series):
    """Remove duplicate rows or values by keeping the first of each duplicate.

    Parameters
    ----------
    df_or_series : :any:`pandas.DataFrame` or :any:`pandas.Series`
        Pandas object from which to drop duplicate index values.

    Returns
    -------
    deduplicated : :any:`pandas.DataFrame` or :any:`pandas.Series`
        The deduplicated pandas object.
    """
    # CalTrack 2.3.2.2
    return df_or_series[~df_or_series.index.duplicated(keep="first")]


def day_counts(index):
    """Days between DatetimeIndex values as a :any:`pandas.Series`.

    Parameters
    ----------
    index : :any:`pandas.DatetimeIndex`
        The index for which to get day counts.

    Returns
    -------
    day_counts : :any:`pandas.Series`
        A :any:`pandas.Series` with counts of days between periods. Counts are
        given on start dates of periods.
    """
    # dont affect the original data
    index = index.copy()

    if len(index) == 0:
        return pd.Series([], index=index)

    timedeltas = (index[1:] - index[:-1]).append(pd.TimedeltaIndex([pd.NaT]))
    timedelta_days = timedeltas.total_seconds() / (60 * 60 * 24)

    return pd.Series(timedelta_days, index=index)


def clean_billing_data(data, source_interval, warnings):
    # check for empty data
    if data["value"].dropna().empty:
        return data[:0]

    if source_interval.startswith("billing"):
        diff = list((data.index[1:] - data.index[:-1]).days)
        filter_ = pd.Series(diff + [np.nan], index=data.index)

        # CalTRACK 2.2.3.4, 2.2.3.5
        if source_interval == "billing_monthly":
            data = data[
                (filter_ <= 35) & (filter_ >= 25)  # keep these, inclusive
            ].reindex(data.index)

            if len(data[(filter_ > 35) | (filter_ < 25)]) > 0:
                warnings.append(
                    EEMeterWarning(
                        qualified_name="eemeter.sufficiency_criteria.offcycle_reads_in_billing_monthly_data",
                        description=(
                            "Off-cycle reads found in billing monthly data having a duration of less than 25 days"
                        ),
                        data=[
                            timestamp.isoformat()
                            for timestamp in data[(filter_ > 35) | (filter_ < 25)].index
                        ],
                    )
                )

        # CalTRACK 2.2.3.4, 2.2.3.5
        if source_interval == "billing_bimonthly":
            data = data[
                (filter_ <= 70) & (filter_ >= 25)  # keep these, inclusive
            ].reindex(data.index)

            if len(data[(filter_ > 70) | (filter_ < 25)]) > 0:
                warnings.append(
                    EEMeterWarning(
                        qualified_name="eemeter.sufficiency_criteria.offcycle_reads_in_billing_monthly_data",
                        description=(
                            "Off-cycle reads found in billing monthly data having a duration of less than 25 days"
                        ),
                        data=[
                            timestamp.isoformat()
                            for timestamp in data[(filter_ > 70) | (filter_ < 25)].index
                        ],
                    )
                )

        # CalTRACK 2.2.3.1
        """
        Adds estimate to subsequent read if there aren't more than one estimate in a row
        and then removes the estimated row.

        Input:
        index   value   estimated
        1       2       False
        2       3       False
        3       5       True
        4       4       False
        5       6       True
        6       3       True
        7       4       False
        8       NaN     NaN

        Output:
        index   value
        1       2
        2       3
        4       9
        5       NaN
        7       7
        8       NaN
        """
        add_estimated = []
        remove_estimated_fixed_rows = []
        orig_data = data.copy()
        if "estimated" in data.columns:
            data["unestimated_value"] = (
                data[:-1].value[(data[:-1].estimated == False)].reindex(data.index)
            )
            data["estimated_value"] = (
                data[:-1].value[(data[:-1].estimated)].reindex(data.index)
            )
            for i, (index, row) in enumerate(data[:-1].iterrows()):
                # ensures there is a prev_row and previous row value is null
                if i > 0 and pd.isnull(prev_row["unestimated_value"]):
                    # current row value is not null
                    add_estimated.append(prev_row["estimated_value"])
                    if not pd.isnull(row["unestimated_value"]):
                        # get all rows that had only estimated reads that will be
                        # added to the subsequent row meaning this row
                        # needs to be removed
                        remove_estimated_fixed_rows.append(prev_index)
                else:
                    add_estimated.append(0)
                prev_row = row
                prev_index = index
            add_estimated.append(np.nan)
            data["value"] = data["unestimated_value"] + add_estimated
            data = data[~data.index.isin(remove_estimated_fixed_rows)]
            data = data[["value"]]  # remove the estimated column

    # check again for empty data
    if data.dropna().empty:
        return data[:0]

    return data["value"].to_frame()


def as_freq(
    data_series,
    freq,
    atomic_freq="1 Min",
    series_type="cumulative",
    include_coverage=False,
):
    """Resample data to a different frequency.

    This method can be used to upsample or downsample meter data. The
    assumption it makes to do so is that meter data is constant and averaged
    over the given periods. For instance, to convert billing-period data to
    daily data, this method first upsamples to the atomic frequency
    (1 minute freqency, by default), "spreading" usage evenly across all
    minutes in each period. Then it downsamples to hourly frequency and
    returns that result. With instantaneous series, the data is copied to all
    contiguous time intervals and the mean over `freq` is returned.

    **Caveats**:

     - This method gives a fair amount of flexibility in
       resampling as long as you are OK with the assumption that usage is
       constant over the period (this assumption is generally broken in
       observed data at large enough frequencies, so this caveat should not be
       taken lightly).

    Parameters
    ----------
    data_series : :any:`pandas.Series`
        Data to resample. Should have a :any:`pandas.DatetimeIndex`.
    freq : :any:`str`
        The frequency to resample to. This should be given in a form recognized
        by the :any:`pandas.Series.resample` method.
    atomic_freq : :any:`str`, optional
        The "atomic" frequency of the intermediate data form. This can be
        adjusted to a higher atomic frequency to increase speed or memory
        performance.
    series_type : :any:`str`, {'cumulative', ‘instantaneous’},
        default 'cumulative'
        Type of data sampling. 'cumulative' data can be spread over smaller
        time intervals and is aggregated using addition (e.g. meter data).
        'instantaneous' data is copied (not spread) over smaller time intervals
        and is aggregated by averaging (e.g. weather data).
    include_coverage: :any:`bool`,
        default `False`
        Option of whether to return a series with just the resampled values
        or a dataframe with a column that includes percent coverage of source data
        used for each sample.

    Returns
    -------
    resampled_data : :any:`pandas.Series` or :any:`pandas.DataFrame`
        Data resampled to the given frequency (optionally as a dataframe with a coverage column if `include_coverage` is used.
    """
    # TODO(philngo): make sure this complies with CalTRACK 2.2.2.1
    if not isinstance(data_series, pd.Series):
        raise ValueError(
            "expected series, got object with class {}".format(data_series.__class__)
        )
    if data_series.empty:
        return data_series
    series = remove_duplicates(data_series)
    target_freq = pd.Timedelta(atomic_freq)
    timedeltas = (series.index[1:] - series.index[:-1]).append(
        pd.TimedeltaIndex([pd.NaT])
    )

    if series_type == "cumulative":
        spread_factor = target_freq.total_seconds() / timedeltas.total_seconds()
        series_spread = series * spread_factor
        atomic_series = series_spread.asfreq(atomic_freq, method="ffill")
        resampled = atomic_series.resample(freq, origin=series.index[0]).sum()
        resampled_with_nans = atomic_series.resample(
            freq, origin=series.index[0]
        ).first()
        n_coverage = atomic_series.resample(freq, origin=series.index[0]).count()
        resampled = resampled[resampled_with_nans.notnull()].reindex(resampled.index)

    elif series_type == "instantaneous":
        # ffill on series.asfreq can produce unintuitive results if resampling a sparse matrix.
        # for example, attempting to resample 2 months of hourly data to daily with a month of
        # absent rows (not NaN, but missing from the dataframe) will ffill that month with the previous read.
        #
        # a similar effect can happen if you have NaNs at a different frequency appended to the end
        # of a series. this could happen if you concat a monthly series with an hourly one at an offset.
        # the call to asfreq() could erroneously fill in a month of data, followed by NaNs
        atomic_series = series.asfreq(atomic_freq, method="ffill")
        resampled = atomic_series.resample(freq, origin=series.index[0]).mean()
        n_coverage = atomic_series.resample(freq, origin=series.index[0]).count()

    # Edit : Added a check so that hourly and daily frequencies don't have a null value at the end
    if freq not in ["h", "D"] and resampled.index[-1] < series.index[-1]:
        # this adds a null at the end using the target frequency
        last_index = pd.date_range(resampled.index[-1], freq=freq, periods=2)[1:]
        resampled = (
            pd.concat([resampled, pd.Series(np.nan, index=last_index)])
            .resample(freq)
            .mean()
        )
    if include_coverage:
        n_total = (
            resampled.resample(atomic_freq)
            .count()
            .resample(freq, origin=resampled.index[0])
            .count()
        )
        resampled = resampled.to_frame("value")
        resampled["coverage"] = n_coverage / n_total

        # TODO : hacky fix to account all occurences of last hour not being counted due to the NaN appended above.
        # Due to above issue number of median granularity periods would end up being 1 rather than the entire 720(24 * 30), thus squashing the
        # reported value to 1/720th the actual. Set it back to 1 like the other usual periods. Might break if the last period is uneven.
        if resampled.coverage.iloc[-1] > 1:
            resampled.iloc[-1, resampled.columns.get_loc("coverage")] = 1
        return resampled
    else:
        return resampled


def downsample_and_clean_daily_data(dataset, warnings):
    dataset = as_freq(dataset, "D", include_coverage=True)

    if not dataset[dataset.coverage <= 0.5].empty:
        warnings.append(
            EEMeterWarning(
                qualified_name="eemeter.sufficiency_criteria.missing_high_frequency_meter_data",
                description=(
                    "More than 50% of the high frequency Meter data is missing."
                ),
                data=[
                    timestamp.isoformat()
                    for timestamp in dataset[dataset.coverage <= 0.5].index
                ],
            )
        )

    # CalTRACK 2.2.2.1 - interpolate with average of non-null values
    dataset.loc[dataset.coverage > 0.5, "value"] = (
        dataset[dataset.coverage > 0.5].value / dataset[dataset.coverage > 0.5].coverage
    )

    return dataset[dataset.coverage > 0.5].reindex(dataset.index)[["value"]]


def clean_billing_daily_data(data, source_interval, warnings):
    # billing data is cleaned but not resampled
    if source_interval.startswith("billing"):
        # CalTRACK 2.2.3.4, 2.2.3.5
        return clean_billing_data(data, source_interval, warnings)

    # higher intervals like daily, hourly, 30min, 15min are
    # resampled (daily) or downsampled (hourly, 30min, 15min)
    elif source_interval == "daily":
        return data.to_frame("value")
    else:
        return downsample_and_clean_daily_data(data, warnings)


# TODO : requires more testing
def compute_minimum_granularity(index: pd.Series, default_granularity: Optional[str]):
    if len(index) <= 1:
        return default_granularity
    # Inferred frequency returns None if frequency can't be autodetected
    index.freq = index.inferred_freq
    if index.freq is None:
        # max_difference = day_counts(index).max()
        # min_difference = day_counts(index).min()
        median_difference = day_counts(index).median()
        # if max_difference == 1 and min_difference == 1:
        #     min_granularity = 'daily'
        # elif max_difference < 1:
        #     min_granularity = 'hourly'
        # elif max_difference >= 60:
        #     min_granularity = 'billing_bimonthly'
        # elif max_difference >= 30:
        #     min_granularity = 'billing_monthly'
        # else:
        #     min_granularity = default_granularity

        granularity_dict = {
            median_difference < 1: "hourly",
            median_difference == 1: "daily",
            1 < median_difference <= 35: "billing_monthly",
            35 < median_difference <= 70: "billing_bimonthly",
        }
        min_granularity = granularity_dict.get(True, default_granularity)
        return min_granularity
    # The other cases still result in granularity being unknown so this causes the frequency to be resampled to daily
    if isinstance(index.freq, MonthEnd) or isinstance(
        index.freq, MonthBegin
    ):  # Can be MonthEnd or MonthBegin instance
        if index.freq.n == 1:
            min_granularity = "billing_monthly"
        else:
            min_granularity = "billing_bimonthly"
    elif index.freq <= pd.Timedelta(hours=1):
        min_granularity = "hourly"
    elif index.freq <= pd.Timedelta(days=1):
        min_granularity = "daily"
    elif index.freq <= pd.Timedelta(days=30):
        min_granularity = "billing_monthly"
    else:
        min_granularity = "billing_bimonthly"

    return min_granularity



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import pytest
from opendsm.eemeter.models.daily.optimize import obj_fcn_dec, Optimizer

from opendsm.eemeter.models.daily.objective_function import obj_fcn_decorator

from opendsm.eemeter.models.daily.base_models.hdd_tidd_cdd import (
    evaluate_hdd_tidd_cdd_smooth,
    _hdd_tidd_cdd_smooth_weight,
)

from opendsm.eemeter.models.daily.fit_base_models import _get_opt_settings
from opendsm.eemeter.models.daily.utilities.settings import DailySettings as Settings


def test_obj_fcn_dec():
    # Test case 1: Check if the function returns the expected output for valid input
    def obj_fcn(x):
        return np.sum(x**2)

    x0 = np.array([1, 2, 3])
    bnds = np.array([[0, 1], [1, 2], [2, 3]])
    obj_fcn_eval, idx_opt = obj_fcn_dec(obj_fcn, x0, bnds)

    x = np.array([0.5, 1.5, 2.5])
    expected_output = 5
    assert obj_fcn_eval(x) == expected_output
    assert idx_opt == [0, 1, 2]


def get_obj_fcn(settings):
    # Test case 1: Test with minimum input values
    model_fcn_full = evaluate_hdd_tidd_cdd_smooth
    weight_fcn = _hdd_tidd_cdd_smooth_weight
    TSS_fcn = None
    T = np.array([1, 2, 3, 4, 5, 6, 7]).astype(float)
    obs = np.array([2, 4, 6, 8, 10, 12, 14]).astype(float)
    base_weights = None
    alpha = 2.0
    coef_id = [
        "hdd_bp",
        "hdd_beta",
        "hdd_k",
        "cdd_bp",
        "cdd_beta",
        "cdd_k",
        "intercept",
    ]
    initial_fit = True
    return obj_fcn_decorator(
        model_fcn_full,
        weight_fcn,
        TSS_fcn,
        T,
        obs,
        base_weights,
        settings,
        alpha,
        coef_id,
        initial_fit,
    )


@pytest.fixture
def get_x0():
    return np.array([73.48349431, 0.0, 0.0, 81.66823342, 5.34878023, 0.0, 50.85274713])


@pytest.fixture
def get_bnds():
    return np.array(
        [
            [59.79057581, 86.91811908],
            [0.0, 16.0463407],
            [0.0, 1.0],
            [59.79057581, 86.91811908],
            [0.0, 16.0463407],
            [0.0, 1.0],
            [20.13393783, 79.33486982],
        ]
    )


def test_optimizer_run(get_x0, get_bnds):
    x0 = get_x0
    bnds = get_bnds
    coef_id = [
        "hdd_bp",
        "hdd_beta",
        "hdd_k",
        "cdd_bp",
        "cdd_beta",
        "cdd_k",
        "intercept",
    ]

    # Test case 1: Test with empty options
    settings = Settings()
    opt_settings = _get_opt_settings(settings)
    obj_fcn = get_obj_fcn(settings)
    optimizer = Optimizer(obj_fcn, x0, bnds, coef_id, settings, opt_settings)
    res = optimizer.run()
    assert np.allclose(res.x, np.array([20.13393]), rtol=1e-5, atol=1e-5)

    # Test case 2: Test with scipy algorithm
    settings = Settings(developer_mode=True, algorithm_choice="scipy_Nelder-Mead")
    opt_settings = _get_opt_settings(settings)
    obj_fcn = get_obj_fcn(settings)
    optimizer = Optimizer(obj_fcn, x0, bnds, coef_id, settings, opt_settings)
    res = optimizer.run()
    assert np.allclose(res.x, np.array([20.13393]), rtol=1e-5, atol=1e-5)

    # Test case 3: Test with nlopt algorithm
    settings = Settings(developer_mode=True, algorithm_choice="nlopt_sbplx")
    opt_settings = _get_opt_settings(settings)
    obj_fcn = get_obj_fcn(settings)
    optimizer = Optimizer(obj_fcn, x0, bnds, coef_id, settings, opt_settings)
    res = optimizer.run()
    assert np.allclose(res.x, np.array([20.13393]), rtol=1e-5, atol=1e-5)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from datetime import datetime

from opendsm.eemeter import (
    HourlyBaselineData,
    HourlyReportingData,
    HourlyModel,
    HourlySolarSettings,
    HourlyNonSolarSettings,
)
from opendsm.eemeter.models.hourly.settings import BaseHourlySettings
from opendsm.eemeter.common.exceptions import (
    DataSufficiencyError,
    DisqualifiedModelError,
)
from opendsm.eemeter.common.warnings import EEMeterWarning
from opendsm.common.test_data import load_test_data
import numpy as np
import pandas as pd
import pytest
from math import ceil

_TEST_METER = 110596


@pytest.fixture
def hourly_data():
    baseline, reporting = load_test_data("hourly_treatment_data")
    return baseline.loc[_TEST_METER], reporting.loc[_TEST_METER]


@pytest.fixture
def baseline(hourly_data):
    baseline, _ = hourly_data
    baseline.loc[baseline["observed"] > 513, "observed"] = (
        0  # quick extreme value removal
    )
    baseline["ghi"] = (
        np.sin(np.linspace(0, 2 * np.pi * len(baseline), len(baseline))) * 40
    ).round(2) + 40
    return baseline


@pytest.fixture
def reporting(hourly_data):
    _, reporting = hourly_data
    reporting["ghi"] = (
        np.sin(np.linspace(0, 2 * np.pi * len(reporting), len(reporting))) * 40
    ).round(2) + 40
    return reporting


@pytest.fixture
def baseline_ghi(baseline):
    # generate ghi as a sin wave daily period peaking afternoon
    baseline["ghi"] = np.sin(np.linspace(0, 2 * np.pi, len(baseline)))
    return baseline


def test_good_data(baseline, reporting):
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    reporting_data = HourlyReportingData(reporting, is_electricity_data=True)
    hm = HourlyModel().fit(baseline_data)
    p1 = hm.predict(reporting_data)
    assert np.isclose(
        p1["predicted"].sum(), 1135000, rtol=1e-2
    )  # quick check that model fit isn't changing drastically
    serialized = hm.to_json()
    hm2 = HourlyModel.from_json(serialized)
    p2 = hm2.predict(reporting_data)
    assert p1.equals(p2)


def test_misaligned_data(baseline, reporting):
    reporting.index = reporting.index.shift(8, freq="h")
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    reporting_data = HourlyReportingData(reporting, is_electricity_data=True)
    hm = HourlyModel().fit(baseline_data)
    hm.predict(reporting_data)


def test_tz_naive(baseline):
    baseline.index = baseline.index.tz_localize(None)
    with pytest.raises(ValueError):
        HourlyBaselineData(baseline, is_electricity_data=True)


def test_tz_mismatch(baseline):
    # might allow automatic adjustment from the model in the future, but hard requirement for now
    baseline.index = baseline.index.tz_convert("US/Pacific")
    reporting = baseline.copy()
    reporting.index = reporting.index.tz_convert("US/Eastern")
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    reporting_data = HourlyReportingData(reporting, is_electricity_data=True)
    hm = HourlyModel().fit(baseline_data)
    with pytest.raises(ValueError):
        hm.predict(reporting_data)


def test_predict_missing_fit_features(baseline, reporting):
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    hm = HourlyModel(settings=HourlySolarSettings()).fit(baseline_data)
    reporting.drop("ghi", axis=1, inplace=True)
    reporting_data = HourlyReportingData(reporting, is_electricity_data=True)
    with pytest.raises(ValueError):
        hm.predict(reporting_data)


def test_nonsolar_predict_with_ghi(baseline, reporting, caplog):
    baseline.drop("ghi", axis=1, inplace=True)
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    hm = HourlyModel().fit(baseline_data)
    reporting_data = HourlyReportingData(reporting, is_electricity_data=True)
    with caplog.at_level("WARNING"):
        hm.predict(reporting_data)
        assert "GHI" in caplog.text


def test_forced_solar_model_fit_no_ghi(baseline):
    baseline = baseline.drop("ghi", axis=1)
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    with pytest.raises(ValueError):
        HourlyModel(settings=HourlySolarSettings()).fit(baseline_data)


def test_forced_nonsolar_model_fit_with_ghi(baseline):
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    hm = HourlyModel(settings=HourlyNonSolarSettings()).fit(baseline_data)
    assert [
        w for w in hm.warnings if w.qualified_name == "eemeter.potential_model_mismatch"
    ]


def test_no_data(baseline):
    baseline["observed"] = 0
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)


def test_negative_meter_values(baseline):
    baseline.loc["2018-01-08", "observed"] = -1

    # gas data can't be negative
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=False)
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)

    # elec can
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    HourlyModel().fit(baseline_data)


def test_invalid_baseline_lengths(baseline):
    # TODO import min/max length from constants
    MAX_BASELINE_HOURS = 8760
    MIN_BASELINE_HOURS = ceil(MAX_BASELINE_HOURS * 0.9) - 24
    short_df = baseline.iloc[:MIN_BASELINE_HOURS]

    extra_day = baseline.iloc[-24:]
    extra_day.index += pd.Timedelta(days=1)
    long_df = pd.concat([baseline, extra_day])

    short_baseline = HourlyBaselineData(short_df, is_electricity_data=True)
    long_baseline = HourlyBaselineData(long_df, is_electricity_data=True)
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(short_baseline)
    hm_short = HourlyModel().fit(short_baseline, ignore_disqualification=True)
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(long_baseline)
    hm_long = HourlyModel().fit(long_baseline, ignore_disqualification=True)


def test_low_freq_temp(baseline):
    baseline["temperature"] = baseline["temperature"].resample("D").mean()
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    assert_dq(
        baseline_data,
        ["eemeter.sufficiency_criteria.too_many_days_with_missing_temperature_data"],
    )
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)


def test_low_freq_meter(baseline):
    baseline["observed"] = baseline["observed"].resample("D").mean()
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    assert_dq(
        baseline_data,
        ["eemeter.sufficiency_criteria.too_many_days_with_missing_meter_data"],
    )
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)


def test_monthly_percentage(baseline):
    missing_idx = pd.date_range(
        start=baseline.index.min(), end=baseline.index.max(), freq="h"
    )
    # create datetimeindex where a little over 10% of days are missing in feb, but still 90% overall
    missing_idx = missing_idx[missing_idx.day < 4]
    invalid_baseline = baseline[~baseline.index.isin(missing_idx)]
    # create datetimeindex where a little under 10% of days are missing in feb
    missing_idx = missing_idx[missing_idx.day < 3]
    valid_baseline = baseline[~baseline.index.isin(missing_idx)]

    invalid_temp = baseline.copy()
    invalid_temp.loc[invalid_temp.index.day < 5, "temperature"] = np.nan

    invalid_meter = baseline.copy()
    invalid_meter.loc[invalid_meter.index.day < 5, "observed"] = np.nan

    baseline_data = HourlyBaselineData(invalid_baseline, is_electricity_data=True)
    assert_dq(
        baseline_data, ["eemeter.sufficiency_criteria.missing_monthly_temperature_data"]
    )
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)
    baseline_data = HourlyBaselineData(valid_baseline, is_electricity_data=True)
    HourlyModel().fit(baseline_data)

    baseline_data = HourlyBaselineData(invalid_temp, is_electricity_data=True)
    assert_dq(
        baseline_data,
        [
            "eemeter.sufficiency_criteria.too_many_days_with_missing_data",
            "eemeter.sufficiency_criteria.missing_monthly_temperature_data",
            "eemeter.sufficiency_criteria.too_many_days_with_missing_temperature_data",
        ],
    )
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)

    baseline_data = HourlyBaselineData(invalid_meter, is_electricity_data=True)
    assert_dq(
        baseline_data,
        [
            "eemeter.sufficiency_criteria.too_many_days_with_missing_data",
            "eemeter.sufficiency_criteria.missing_monthly_meter_data",
            "eemeter.sufficiency_criteria.too_many_days_with_missing_meter_data",
        ],
    )
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)


def test_monthly_ghi_percentage(baseline):
    # create datetimeindex where a little over 10% of days are missing in feb, but still 90% overall
    missing_idx = pd.date_range(
        start=baseline.index.min(), end=baseline.index.max(), freq="h"
    )
    missing_idx = missing_idx[missing_idx.day < 4]

    invalid_ghi = baseline.copy()
    invalid_ghi.loc[invalid_ghi.index.day < 5, "ghi"] = np.nan

    baseline_data = HourlyBaselineData(invalid_ghi, is_electricity_data=True)
    assert_dq(
        baseline_data,
        [
            "eemeter.sufficiency_criteria.missing_monthly_ghi_data",
        ],
    )
    with pytest.raises(DataSufficiencyError):
        HourlyModel().fit(baseline_data)


def test_hourly_fit_daily_threshold(baseline):
    """confirm that days with >50% interpolated data are excluded from fit step"""

    # bit fragile testing private methods this way, but fine for now
    m = HourlyModel()
    b1 = baseline.copy()
    b1.loc["2018-01-08":"2018-01-08 11", "temperature"] = np.nan
    b1 = m._add_categorical_features(b1)
    b1 = m._daily_fitting_sufficiency(b1)
    assert b1.loc["2018-01-08", "include_date"].sum() == 24

    b2 = baseline.copy()
    b2.loc["2018-01-08":"2018-01-08 12", "temperature"] = np.nan
    b2 = m._add_categorical_features(b2)
    b2 = m._daily_fitting_sufficiency(b2)
    assert b2.loc["2018-01-08", "include_date"].sum() == 0
    assert b2.loc["2018-01-09", "include_date"].sum() == 24


@pytest.mark.filterwarnings("ignore:Objective did not converge.")
def test_hourly_error_metric_dq(baseline):
    baseline["observed"] = np.random.normal(-1, 10, len(baseline)) ** 3
    baseline_data = HourlyBaselineData(baseline, is_electricity_data=True)
    model = HourlyModel().fit(baseline_data)
    assert_dq(baseline_data, ["eemeter.model_fit_metrics"])
    with pytest.raises(DisqualifiedModelError):
        model.predict(baseline_data)


def assert_dq(data, expected_disqualifications):
    remaining_dq = set(expected_disqualifications)
    for dq in data.disqualification:
        if dq.qualified_name in remaining_dq:
            remaining_dq.remove(dq.qualified_name)
    assert not remaining_dq


def test_hourly_dict_settings():
    m = HourlyModel(settings={"train_features": ["feature_col"]})
    assert isinstance(m.settings, HourlyNonSolarSettings)
    assert set(m.settings.train_features) == {"temperature", "feature_col"}
    m = HourlyModel(settings={"train_features": ["feature_col", "ghi"]})
    assert isinstance(m.settings, HourlySolarSettings)
    assert set(m.settings.train_features) == {"temperature", "ghi", "feature_col"}
    m = HourlyModel(settings={"cvrmse_threshold": 1.0})
    assert isinstance(m.settings, BaseHourlySettings)
    assert m.settings.train_features == None
